#!/usr/bin/perl

# Fast File Search crawler
# Copyright (c) 2002-2005 Josef Zlomek
#
# This file is part of Fast File Search (FFS).
#
# FFS is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA,
# or download it from http://www.gnu.org/licenses/gpl.html
#
# FFS uses some code from FemFind by Martin Richtarsky
# (http://femfind.sourceforge.net)

use strict;
use DBI;
use Getopt::Long;
use Net::FTP;
use String::CRC32;
use Fcntl;
use POSIX qw(tmpnam);
use IPC::SysV qw(IPC_CREAT IPC_PRIVATE SEM_UNDO S_IRWXU);
use IPC::Semaphore;
use BSD::Resource;

# include the common package
$_ = $0;
$_ =~ s/[^\/]*$/common.pl/;
require;

my (%c, @arg);
my ($mdb, $db, $sem);
my $file_listing_size;

# global variables for FTP crawling
my ($ftp_handle, $ftp_host, $ftp_ip, $ftp_port, $ftp_passive, $ftp_ls_recursive);
my ($ftp_login, $ftp_password, $ftp_cur_dir);

# read configuration and process command line options
&read_config (\%c);

my ($help, $operation, $lookup, $no_lock, $no_smb, $no_ftp);
if (&process_options ())
{
	exit 1;
}

# get semaphores
$sem = new IPC::Semaphore (IPC_PRIVATE, 2, S_IRWXU | IPC_CREAT);
if (!defined ($sem))
{
	print "Can't get semaphores, reason: $!\n";
	exit 1;
}
$sem->setall ((1) x 2);

# open log file
open (LOG, ">>$c{LOG_FILE}") || die "Could not open $c{LOG_FILE}\nReason: $!\n";
my $tmp = select (LOG); $| = 1; select ($tmp);

# start crawl
$SIG{PIPE} = "IGNORE";
my $starting_time = time ();
my $cmd_line = join (' ', $0, @ARGV);
print LOG '-' x 79, "\n", scalar localtime ($starting_time), " starting $cmd_line\n", '-' x 79, "\n";
my $ret_value = &run ();
my $time_elapsed = time () - $starting_time;
print LOG '-' x 79, "\n", scalar localtime (), " finished, time elapsed: ";
printf LOG "%02d:%02d:%02d\n", $time_elapsed / 3600, ($time_elapsed / 60) % 60, $time_elapsed % 60;

# cleanup
close (LOG);
$sem->remove ();

exit $ret_value;

######################### SUBROUTINES #########################################

# process command line options and set corresponding global variables
sub process_options
{
	# number the various operations and define symbolic names
	my $arg_n = 1;
	$c{CRAWL_ALL} = $arg_n++;
	$c{CRAWL_INC} = $arg_n++;
	$c{CRAWL_FORCE} = $arg_n++;
	$c{OPTIMIZE_TABLES} = $arg_n++;
	$c{CREATE_TABLES} = $arg_n++;
	$c{FIX_TABLES} = $arg_n++;
	$c{REPAIR_TABLES} = $arg_n++;
	$c{DELETE_HOSTS} = $arg_n++;
	$c{DELETE_HIDS} = $arg_n++;

	# process the options
	Getopt::Long::Configure ("bundling", "no_ignore_case");
	GetOptions ('help|h' => \$help,
		    'complete|c' => sub { $operation = $c{CRAWL_ALL} },
		    'incremental|i' => sub { $operation = $c{CRAWL_INC} },
		    'force|f' => sub { $operation = $c{CRAWL_FORCE} },
		    'optimize|o' => sub { $operation = $c{OPTIMIZE_TABLES} },
		    'tables|t' => sub { $operation = $c{CREATE_TABLES} },
		    'fix|r' => sub { $operation = $c{FIX_TABLES} },
		    'repair|R' => sub { $operation = $c{REPAIR_TABLES} },
		    'delete-hosts|d' => sub { $operation = $c{DELETE_HOSTS} },
		    'delete-hids|D' => sub { $operation = $c{DELETE_HIDS} },
		    'lookup|l' => \$lookup,
		    'no-lock' => \$no_lock,
		    'no-smb' => \$no_smb,
		    'no-ftp' => \$no_ftp,
		   );
	if ($help)
	{
		&help ();
		exit 0;
	}
	if ($operation == 0)
	{
		&help ();
		return 1;
	}
	return 0;
}

# display the usage of ffsearch.pl
sub help
{
	print "Fast File Search crawler $c{VERSION}\n";
	print "Usage: ffsearch.pl [OPTION]... OPERATION [LIST]\n\n";
	print "Operations\n";
	print "    -f, --force           force crawling of hosts\n";
	print "                            If LIST of host names is specified,\n";
	print "                            only these hosts will be crawled.\n";
	print "    -c, --complete        complete crawl\n";
	print "                            If LIST of host names is specified,\n";
	print "                            only these hosts will be crawled.\n";
	print "    -i, --incremental     incremental (i.e. subsequent) crawl\n";
	print "                            If LIST of host names is specified,\n";
	print "                            only these hosts will be crawled.\n";
	print "    -o, --optimize        optimize tables\n";
	print "    -t, --tables          create tables (old data will be lost!)\n";
	print "    -r, --fix             fix inconsistencies in tables\n";
	print "    -R, --repair          repair & fix tables\n";
	print "    -d, --delete-hosts    delete host\n";
	print "                            LIST is the list of host names,\n";
	print "                            no LIST means all hosts.\n";
	print "    -D, --delete-hids     delete hosts\n";
	print "                            LIST is the list of host IDs,\n";
	print "                            no LIST means all hosts.\n";
	print "\nOptions\n";
	print "        --lookup          lookup the names from LIST by nmblookup\n";
	print "        --no-lock         do not protect multiple ffsearch.pl scripts\n";
	print "                            from running in the same time\n";
	print "        --no-smb          do not crawl SMB servers\n";
	print "        --no-ftp          do not crawl FTP servers\n";
}

# print a message to log
# \param $string Message to print.
# \param $level Debug level to print the message to.
sub log_print
{
	my ($string, $level) = @_;
	chomp ($string);

	if ($level < 0)
	{
		&sem_down ($c{SEM_LOG});
		print LOG "$string\n";
		&sem_up ($c{SEM_LOG});
	}
	elsif ($level <= $c{DEBUG_LEVEL})
	{
		&sem_down ($c{SEM_LOG});
		print LOG "<$level> [$$] $string\n";
		&sem_up ($c{SEM_LOG});
	}
}

# lock the execution, i.e. allow only one instance to be running at a time
sub lock_process
{
	# test whether the lock file exists
	if (open (IN, "<$c{LOCK_FILE}"))
	{
		&log_print ("Lockfile from a previous instance exists, checking if process is still running", 0);
		my $pid = int (<IN>);
		close (IN);

		# check whether the process is still running
		# or the lock file is orphan
		my @out = `ps -p $pid`;
		foreach (@out)
		{
			if (/\s*\Q$pid\E\s+.*ffsearch/)
			{
				&log_print ("Previous instance still running. you might want to adjust your crontab settings. exit\n", 0);
				return 0;
			}
			&log_print ("process not running anymore. continuing", 0);
		}
	}

	# create the lock file
	if (!open (FO, ">$c{LOCK_FILE}"))
	{
		&log_print ("Could not create lock file!", 0);
		return 0;
	}
	print FO $$;
	close FO;

	return 1;
}

# unlock the execution of this script
sub unlock_process
{
	unlink ($c{LOCK_FILE});
}

# run the wanted action
sub run
{
	if (!$no_lock)
	{
		if (!&lock_process ())
		{
			return 0;
		}
	}

	&db_connect (\%c, \$mdb, \$db, \&log_print);
	if (!$db)
	{
		&unlock_process () if (!$no_lock);
		return 1;
	}

	if ($operation == $c{CREATE_TABLES})
	{
		&log_print ("Creating tables.", 1);
		&create_tables ($db);
	}
	elsif ($operation == $c{FIX_TABLES})
	{
		&fix_tables ();
		&update_status ();
	}
	elsif ($operation == $c{REPAIR_TABLES})
	{
		&log_print ("Repairing tables.", 1);
		&repair_tables ();
		&fix_tables ();
		&update_status ();
	}
	elsif ($operation == $c{OPTIMIZE_TABLES})
	{
		&optimize_tables ();
	}
	elsif ($operation == $c{DELETE_HOSTS})
	{
		&log_print ("Deleting hosts.", 1);
		my $HIDS;
		if (@ARGV == 0)
		{
			$HIDS = $db->selectcol_arrayref ("SELECT HID FROM host");
			foreach my $HID (@$HIDS)
			{
				&delete_host ($HID);
			}
		}
		else
		{
			my $where = "";
			foreach my $host (@ARGV)
			{
				$where .= " OR" if ($where ne "");
				$where .= " HostName LIKE " . $db->quote ($host);
			}

			$HIDS = $db->selectcol_arrayref ("SELECT HID FROM smb WHERE" . $where);
			foreach my $HID (@$HIDS)
			{
				&delete_host ($HID);
			}

			$HIDS = $db->selectcol_arrayref ("SELECT HID FROM ftp WHERE" . $where);
			foreach my $HID (@$HIDS)
			{
				&delete_host ($HID);
			}
		}
		&update_status ();
	}
	elsif ($operation == $c{DELETE_HIDS})
	{
		&log_print ("Deleting hosts.", 1);
		my $HIDS;
		if (@ARGV == 0)
		{
			my $HIDS = $db->selectcol_arrayref ("SELECT HID FROM host");
			foreach my $HID (@$HIDS)
			{
				&delete_host ($HID);
			}
		}
		else
		{
			foreach my $HID (@ARGV)
			{
				$HID = int ($HID);
				&delete_host ($HID) if ($HID > 0);
			}
		}
		&update_status ();
	}
	else
	{
		&timeout_log ();
		&timeout_history ();
		&fix_incorrect_shutdown ();
		&smb_crawl () if (!$no_smb && $c{SMB_DISABLE} eq 'no');

		# reconnect to database
		&db_disconnect ($mdb, $db);
		&db_connect (\%c, \$mdb, \$db, \&log_print);
		if (!$db)
		{
			&unlock_process () if (!$no_lock);
			return 1;
		}

		&ftp_crawl () if (!$no_ftp && $c{FTP_DISABLE} eq 'no');

		# reconnect to database
		&db_disconnect ($mdb, $db);
		&db_connect (\%c, \$mdb, \$db, \&log_print);
		if (!$db)
		{
			&unlock_process () if (!$no_lock);
			return 1;
		}

		&update_status ();
	}
	&db_disconnect ($mdb, $db);
	&unlock_process () if (!$no_lock);
	return 0;
}

# get current date and store it to configuration array
sub get_time
{
	my @datetime = localtime (time ());
	$c{year} = $datetime[5] + 1900;
	$c{month} = $datetime[4] + 1;
	$c{day} = $datetime[3];
}

# delete old log records
sub timeout_log
{
	&log_print ("Deleting old log records.", 1);

	my $expire_time = $starting_time - $c{TIMEOUT_LOG} * 24 * 3600;
	$db->do ("DELETE FROM log WHERE Time<$expire_time");
}

# delete search history of non-active users
sub timeout_history
{
	&log_print ("Deleting search history for non-active users.", 1);

	my $expire_time = $starting_time - $c{TIMEOUT_HISTORY} * 24 * 3600;
	my $UIDs = $db->selectcol_arrayref ("SELECT UID FROM user WHERE LastSearched<$expire_time");
	foreach (@$UIDs)
	{
		$db->do ("DELETE FROM history WHERE UID=$_");
	}
	$db->do ("DELETE FROM user WHERE LastSearched<$expire_time");
}

# fix database after incorrect shutdown
sub fix_incorrect_shutdown
{
	my $HID = $db->selectrow_array ("SELECT UpdatingHost FROM status");
	if ($HID)
	{
		&log_print ("Crawler was not cleanly shutdown, fixing database", 1);
		&delete_host ($HID);
		$db->do ("UPDATE status SET UpdatingHost=0");
	}
}

# "down" on semaphore, equivalent to "lock"
# \param $num ID of the semaphore (sequential number)
sub sem_down
{
	my ($num) = @_;
	$sem->op ($num, -1, SEM_UNDO);
}

# "up" on semaphore, equivalent to "unlock"
# \param $num ID of the semaphore (sequential number)
sub sem_up
{
	my ($num) = @_;
	$sem->op ($num, 1, SEM_UNDO);
}

# get name of temporary file
sub get_temp_file_name
{
	my $temp;
	do {
		$temp = tmpnam ();
	} until (sysopen (FH, $temp, O_RDWR | O_CREAT | O_EXCL, 0600));
	close (FH);
	return $temp;
}

# crawl SMB servers
sub smb_crawl
{
	my $hosts = &smb_get_host_list ();
	my $running = 0;

	# scan the hosts
	foreach my $host (@$hosts)
	{
		if ($c{SMB_PARALLEL} == 0)
		{
			if (wait () == -1)
			{
				&log_print ("There should be $running child processes but there are none.", 1);
				$c{SMB_PARALLEL} = $running;
				$running = 0;
			}
			else
			{
				$running--;
				$c{SMB_PARALLEL}++;
			}
		}

		my $started = &smb_scan_host ($host->[0], $host->[1], $host->[2], $host->[3]);
		$c{SMB_PARALLEL} -= $started;
		$running += $started;
	}

	# wait for children
	while ($running > 0)
	{
		if (wait () == -1)
		{
			&log_print ("There should be $running child processes but there are none.", 1);
			last;
		}
		$running--;
	}
	&log_print ("SMB crawl finished", 1);
}

# get list of SMB master browsers using smbclient
# \param $host SMB server to ask for the list of master browsers
# \param $workgroup Workgroup used when connection to $host
# \param $masters Reference to array of master browsers
sub smb_get_masters
{
	my ($host, $workgroup, $masters) = @_;
	my ($ht, $wt);

	if ($host eq '')
	{
		$ht = "LOCALHOST";
	}
	else
	{
		$ht = &shell_quote ($host);
	}
	if ($workgroup)
	{
		$wt = " -W " . &shell_quote ($workgroup);
	}

	my @out = `$c{SMB_PATH}/smbclient $c{SMB_CONF} $c{SMB_ACCPARAM} -L $ht$wt 2>/dev/null`;
	shift @out while ($out[0] !~ /^\tWorkgroup/ && @out > 0);
	shift @out;shift @out;

	foreach (@out)
	{
		chomp;
		my $wgroup = substr ($_, 1, 16);
		my $master = substr ($_, 21);
		$wgroup =~ s/\s+$//;
		$master =~ s/^\s+//;
		next if ($wgroup eq '' || $master eq '');
		push (@$masters, [ $master, $wgroup ] );
	}
}

# get list of SMB master browsers using nmblookup
# \param $masters Reference to array of master browsers
sub smb_nmb_get_masters
{
	my ($masters) = @_;

	# get the list of IP addresses
	my @ip;
	my @out = `$c{SMB_PATH}/nmblookup $c{SMB_CONF} -M -- -`;
	foreach (@out)
	{
		if (/^(\d+\.\d+\.\d+\.\d+)/)
		{
			push (@ip, $1) if (!&is_invalid_IP ($1));
		}
	}

	# get name and work group of each master browser
	foreach (@ip)
	{
		@out = `$c{SMB_PATH}/nmblookup $c{SMB_CONF} -A $_`;
		my $master;
		my $wgroup;
		foreach (@out)
		{
			if (/^\t/)
			{
				my $name = substr ($_, 1, 16);
				my $type = substr ($_, 18, 2);
				my $flags = substr ($_, 24);
				$name =~ s/\s+//;
				$master = $name if ($type eq "20" && $flags =~ /ACTIVE/);
				$wgroup = $name if ($type eq "1d" && $flags =~ /ACTIVE/);
			}
		}
		next if ($wgroup eq '' || $master eq '');
		push (@$masters, [ $master, $wgroup ] );
	}
}

# lookup SMB names using nmblookup
# \param $list List of names to lookup
sub smb_lookup_names
{
	my ($list) = @_;

	my @quoted_list = map { &shell_quote ($_) } @$list;
	my @output = `$c{SMB_PATH}/nmblookup $c{SMB_CONF} -- @quoted_list`;
	my @worklist;
	my $i;
	for ($i = 0; $i < @output; $i++)
	{
		if ($output[$i] =~ /^(\d+\.\d+\.\d+\.\d+)/)
		{
			push (@worklist, $1);
		}
	}
	return &smb_lookup_addresses (\@worklist);
}

# get name and workgroup of SMB server on IP address
# \param $list List of IP addresses
sub smb_lookup_addresses
{
	my ($list) = @_;

	my @output = `$c{SMB_PATH}/nmblookup $c{SMB_CONF} -A @$list`;
	my $i;
	my @HIDs;
	for ($i = 0; $i < @output; $i++)
	{
		if ($output[$i] =~ /(\d+\.\d+\.\d+\.\d+)$/)
		{
			my $ip = $1;
			my $host;
			my $group;
			for ($i++; $output[$i] !~ /^\r?\n$/; $i++)
			{
				if (substr ($output[$i], 18, 2) eq "00")
				{
					if (substr ($output[$i], 25, 5) eq "GROUP")
					{
						$group = substr ($output[$i], 1, 15);
						$group =~ s/\s+$//;
					}
					else
					{
						$host = substr ($output[$i], 1, 15);
						$host =~ s/\s+$//;
					}
				}
			}
			if ($host)
			{
				my $HID = &smb_update_host ($host, $group);
				push (@HIDs, $HID);
			}
		}
	}
	return \@HIDs;
}

# Update the record about host in database, this function expects the tables
# host and smb to be locked.
# \param $host SMB server.
# \param $wgroup Work group.
sub smb_update_host
{
	my ($host, $wgroup) = @_;

	my ($HID, $flags) = $mdb->selectrow_array ("SELECT host.HID,Flags FROM smb LEFT JOIN host USING (HID) WHERE HostName=" . $mdb->quote ($host));
	if ($HID)
	{
		if (($flags & $c{HOST_MANUAL_CHANGES}) == 0)
		{
			my $clear_flags = $c{HOST_DELETE} | $c{HOST_PURGE};
			my $str;
			$str = "ExpireCount=1,Counter=Period," if ($flags & ($c{HOST_DELETE} | $c{HOST_PURGE}));
			$mdb->do ("UPDATE host SET ${str}Flags=Flags&~$clear_flags WHERE HID=$HID");
			$mdb->do ("UPDATE smb SET WorkGroup=" . $mdb->quote ($wgroup) . " WHERE HID=$HID");
		}
	}
	else
	{
		$mdb->do ("INSERT INTO host (HostType,ExpireCount,Period,Counter) VALUES ($c{SMB},1,1,1)");
		$mdb->do ("INSERT INTO smb (HID,HostName,WorkGroup) VALUES (LAST_INSERT_ID()," . $mdb->quote ($host) . "," . $mdb->quote ($wgroup) . ")");
		($HID) = $mdb->selectrow_array ("SELECT host.HID FROM smb LEFT JOIN host USING (HID) WHERE HostName=" . $mdb->quote ($host));
	}
	return $HID;
}

# Get list of SMB servers to scan
sub smb_get_host_list
{
	if (! -x "$c{SMB_PATH_UNQUOTED}/smbclient")
	{
		&log_print ("$c{SMB_PATH_UNQUOTED}/smbclient is not an executable file", 0);
		return undef;
	}
	if (! -x "$c{SMB_PATH_UNQUOTED}/nmblookup")
	{
		&log_print ("$c{SMB_PATH_UNQUOTED}/nmblookup is not an executable file", 0);
		return undef;
	}

	my $which = '';
	if (@ARGV == 0)
	{
		&log_print ("Getting list of SMB servers.", 1);
		my @masters;
		&smb_get_masters ($c{SMB_SERVER}, $c{SMB_WORKGROUP}, \@masters);
		if (@masters == 1 && $masters[0][0] ne $c{SMB_SERVER})
		{
			# $c{SMB_SERVER} is not a master browser
			my $master = $masters[0][0];
			my $wgroup = $masters[0][1];
			@masters = ();
			&smb_get_masters ($master, $wgroup, \@masters);
		}
		if (@masters == 0)	# no master browsers are known
		{
			&smb_nmb_get_masters (\@masters);
		}

		foreach (@masters)
		{
			my $mt = &shell_quote ($_->[0]);
			my $wt = &shell_quote ($_->[1]);
			my @out = `$c{SMB_PATH}/smbclient $c{SMB_CONF} $c{SMB_ACCPARAM} -L $mt -W $wt 2>/dev/null`;
			shift @out while ($out[0] !~ /^\tServer/ && @out > 0);
			shift @out;shift @out;
			my $index = 0;
			$index++ while ($out[$index] !~ /\tWorkgroup/ && $index <= $#out);
			next if ($index == $#out);
			$#out = $index - 2;

			my $wgroup;
			if ($c{CONVERT_TO_UPPER_CASE}) {
				$wgroup = uc ($_->[1]);
			} else {
				$wgroup = $_->[1];
			}
			$mdb->do ("LOCK TABLES host WRITE, smb WRITE");
			foreach (@out)
			{
				my $host;
				if ($c{CONVERT_TO_UPPER_CASE}) {
					$host = uc (substr ($_, 1, 16));
				} else {
					$host = substr ($_, 1, 16);
				}
				$host =~ s/\s+$//;
				&smb_update_host ($host, $wgroup);
			}
			$mdb->do ("UNLOCK TABLES");
		}
	}
	else
	{
		if ($lookup)
		{
			my $HIDs = &smb_lookup_names (\@ARGV);
			if (@$HIDs > 0)
			{
				$which = ' AND (host.HID=' . join (' OR host.HID=', @$HIDs) . ')';
			}
			else
			{
				$which = ' AND 0';
			}
		}
	}

	&get_host_list_from_master ($c{SMB});
	&delete_or_purge_marked_hosts ($c{SMB});

	if (@ARGV > 0 && !$lookup)
	{
		foreach (@ARGV)
		{
			my $HIDs = $db->selectcol_arrayref ("SELECT HID FROM smb WHERE HostName LIKE " . $db->quote ($_));
			foreach (@$HIDs)
			{
				$which .= " OR host.HID=$_";
			}
		}
		$which =~ s/^ OR (.*)/ AND ($1)/;
		if ($which eq '')
		{
			$which = ' AND 0';
		}
	}

	if ($operation == $c{CRAWL_FORCE})
	{
		$db->do ("UPDATE host SET Counter=Period WHERE HostType=$c{SMB} AND Counter<Period $which");
		$db->do ("UPDATE host SET ExpireCount=ExpireCount+1 WHERE HostType=$c{SMB} AND Counter>=Period AND Counter>0 AND ExpireCount<255 $which");
	}
	elsif ($operation == $c{CRAWL_ALL})
	{
		$db->do ("UPDATE host SET Counter=Counter+1 WHERE HostType=$c{SMB} AND Counter<Period $which");
		$db->do ("UPDATE host SET ExpireCount=ExpireCount+1 WHERE HostType=$c{SMB} AND Counter>=Period AND Counter>0 AND ExpireCount<255 $which");
	}
	my $unset_flags = $c{HOST_DELETE} | $c{HOST_PURGE};
	my $hosts =  $db->selectall_arrayref ("SELECT host.HID,HostName,WorkGroup,IP,Flags FROM smb LEFT JOIN host USING (HID) WHERE ExpireCount>0 AND Counter>=Period AND Counter>0 AND (Flags&$unset_flags)=0 $which ORDER BY ScanOrder,HostName");
	foreach (@$hosts)
	{
		if (($_->[4] & $c{HOST_MANUAL_CHANGES}) == 0)
		{
			$_->[3] = '';
		}
	}
	return $hosts;
}

# Scan a SMB server
# \param $HID Host ID of the server
# \param $host SMB name of the server
# \param $workgroup Workgroup of the server
# \param $IP IP address of the server
sub smb_scan_host
{
	my ($HID, $host, $workgroup, $IP) = @_;

	$IP = '' if (&is_invalid_IP ($IP));
	my $pid = fork ();
	if (!defined ($pid))
	{
		&log_print ("Could not fork, reason: $!", 0);
		return 0;
	}
	if ($pid == 0)
	{
		&db_connect (\%c, \$mdb, \$db, \&log_print);
		exit 1 if (!$db);

		my $ht = &shell_quote ($host);
		my $wt;
		my $ip = $IP;
		$IP = "-I " . &shell_quote ($IP) if ($IP);
		$wt = "-W " . &shell_quote ($workgroup) if ($workgroup);
		my @out = `$c{SMB_PATH}/smbclient $c{SMB_CONF} $c{SMB_ACCPARAM} -L $ht $wt $IP 2>&1`;
		if ($? != 0)
		{
			while (@out > 0)
			{
				if ($out[0] =~ /Domain=[\[]([^\]]*)[\]]/)
				{
					&log_print ("//$host: could not retrieve share listing. Probably wrong login/password.", 1);
					$db->do ("UPDATE host SET Counter=Period-1 WHERE HID=$HID AND Period>0");
					$db->do ("UPDATE host SET Flags=Flags|$c{HOST_ONLINE},LastScan=NOW() WHERE HID=$HID");
					last;
				}
				shift @out;
			}
			if (@out == 0)
			{
				&log_print ("//$host: could not connect.", 1);
			}
			&sem_down ($c{SEM_DB});
			my $expire = $db->selectrow_array ("SELECT ExpireCount FROM host WHERE HID=$HID");
			if ($expire >= $c{SMB_EXPIRE})
			{
				&log_print ("//$host: expired, deleting.", 1);
				$db->do ("UPDATE host SET Flags=Flags|$c{HOST_DELETE} WHERE HID=$HID AND Flags&$c{HOST_MANUAL_CHANGES}=0");
				$mdb->do ("UPDATE host SET Flags=Flags|$c{HOST_DELETE} WHERE HID=$HID AND Flags&$c{HOST_MANUAL_CHANGES}=0");
				&delete_host ($HID);
			}
			&sem_up ($c{SEM_DB});
			&db_disconnect ($mdb, $db);
			exit 0;
		}
		while ($out[0] !~ /^\tSharename/ && @out > 0)
		{
			if (!$ip && $out[0] =~ /positive name.*?\( (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) \)/)
			{
				$ip = $1 if (!&is_invalid_IP ($1));
			}
			if ($out[0] =~ /Domain=[\[]([^\]]*)[\]]/)
			{
				$workgroup = $1;
			}
			shift @out;
		}
		shift @out;shift @out;

		$ip = &smb_get_ip ($host) if (!$ip);
		$db->do ("UPDATE host SET IP='$ip' WHERE HID=$HID");
		$mdb->do ("UPDATE host SET IP='$ip' WHERE HID=$HID") if ($mdb != $db);
		$db->do ("UPDATE smb SET WorkGroup=" . $db->quote ($workgroup) . " WHERE HID=$HID");
		$mdb->do ("UPDATE smb SET WorkGroup=" . $mdb->quote ($workgroup) . " WHERE HID=$HID") if ($mdb != $db);
		&create_path ($HID, 0, '');

		my (@shares, $share);
		while (@out > 0 && $out[0] ne '')
		{
			$_ = shift (@out);
			if (/^\t(.{15}) ?Disk/)
			{
				if ($c{CONVERT_TO_UPPER_CASE}) {
					$share = uc ($1);
				} else {
					$share = $1;
				}
				$share =~ s/^\s+//;
				$share =~ s/\s+$//;
				next if ($share eq '');
				next if ($share =~ /\$$/ && $c{SMB_CRAWL_HIDDEN} eq 'no');
				next if ($share eq 'IPC$' || $share eq 'ADMIN$');
				push (@shares, $share);
			}
		}

		# crawl shares
		my $temp = &get_temp_file_name ();
		my %share_count;
		foreach $share (@shares)
		{
			my $pid = fork ();
			if (!defined ($pid))
			{
				&log_print ("Could not fork, reason: $!", 0);
			}
			elsif ($pid == 0)
			{
				my $ht = &shell_quote ($host);
				my $st = &shell_quote ($share);
				my $wt;
				$wt = "-W " . &shell_quote ($workgroup) if ($workgroup);
				&log_print ("//$host/$share: crawling", 1);
				setrlimit (RLIMIT_FSIZE, $c{MAX_TEMP_FILE_SIZE}, $c{MAX_TEMP_FILE_SIZE}) if ($c{MAX_TEMP_FILE_SIZE});
				exec ("$c{SMB_PATH}/smbclient //$ht/$st $c{SMB_CONF} $c{SMB_ACCPARAM} -c 'recurse;ls \\*' -d0 -D . $wt $IP >$temp 2>/dev/null");
				exit 1;
			}
			else
			{
				if (wait () != -1 && $? == 0)
				{
					open (IN, "<$temp");
					&smb_update_database ($HID, $host, $share, *IN);
					close (IN);
					$share_count{$share}--;
				}
			}
		}

		&sem_down ($c{SEM_DB});
		my @purge_share;
		my $shares = $db->selectall_arrayref ("SELECT SID,ShareName FROM share WHERE HID=$HID");
		foreach (@$shares)
		{
			$share_count{$_->[1]}++;
			push (@purge_share, $_->[0]) if ($share_count{$_->[1]} == 1);
		}
		if (@purge_share > 0)
		{
			&log_print ("//$host: purging non-accessible shares", 1);
			&purge_shares (\@purge_share);
		}
		&sem_up ($c{SEM_DB});

		&db_disconnect ($mdb, $db);
		unlink ($temp);
		exit 0;
	}
	return 1;
}

# get IP address of SMB server
# \param $host SMB name of the server
sub smb_get_ip
{
	my ($host) = @_;

	return $host if (!&is_invalid_IP ($host));

	my $i = 0;
	my $ip;

	# first try nmblookup
	my $ht = &shell_quote ($host);
	my @out = `$c{SMB_PATH}/nmblookup $c{SMB_CONF} -- $ht`;
	if ($? == 0)
	{
		foreach (@out)
		{
			if (/^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) $host/)
			{
				if (!&is_invalid_IP ($1))
				{
					$i++;
					$ip = $1 if ($i == 1);
				}

				return $1 if (!&is_invalid_IP ($1));
			}
		}
	}
	# return the IP address if there was only one IP address in the output of nmblookup
	return $ip if ($i == 1);

	# then try to resolve it by DNS
	my @addr = (gethostbyname ($host))[4];
	return join ('.', unpack ('C4', $addr[0])) if (@addr);

	# if we do not have the IP address from DNS and have multiple addresses from nmblookup
	# return the first of them
	return $ip if ($i > 0);

	&log_print ("//$host: can't get IP address\n", 1);
	return "";
}

# update the content of SMB server in database 
# \param $HID Host ID of the server
# \param $host SMB name of the server
# \param $share SMB share on the server
# \param $filehandle Opened file handle of file containing the list of files of the server
sub smb_update_database
{
	my ($HID, $host, $share, $filehandle) = @_;

	&sem_down ($c{SEM_DB});
	&log_print ("//$host/$share: updating database", 1);
	$db->do ("UPDATE status SET UpdatingHost=$HID");

	my $select = "SELECT SID FROM share WHERE HID=$HID AND ShareName=" . $db->quote ($share);
	my $SID = $db->selectrow_array ($select);
	if (!defined ($SID))
	{
		$db->do ("INSERT INTO share (HID,ShareName) VALUES ($HID," . $db->quote ($share) . ")");
		$SID = $db->selectrow_array ($select);
	}

	if (eof ($filehandle))
	{
		&purge_share ($SID);
		&sem_up ($c{SEM_DB});
		return;
	}

	my %path_count;
	my (%filedate, %filesize);
	my $next_root = "/$share";
	$_ = <$filehandle>;
	chomp;
	while (!eof ($filehandle))
	{
		my $root = $next_root;
		my $next_dir = 0;

		my @files = ();
		%filedate = ();
		%filesize = ();
		do {
			$_ = <$filehandle>;
			chomp;
			if (/^  /)
			{
				my $fsize = substr ($_, -44, 18);
				if ($fsize =~ /(\d+)$/)
				{
					$fsize = $1;
					my $pos = -34 - length ($fsize);
					$pos = -42 if ($pos > -42);
					my $fattrib = substr ($_, $pos, 7);
					if ($fattrib !~ /D/)
					{
						my $fyear = substr ($_, -4, 4);
						my $ftime = substr ($_, -13, 8);
						my $fmonth = substr ($_, -20, 3);
						my $fday = substr ($_, -16, 2);
						my $fname = substr ($_, 2, length ($_) + $pos);
						$fname =~ s/\s+$//;
						$fday =~ s/^\s+//;
						$fday = "0$fday" if (length ($fday) == 1);
						$fmonth = "Jan" if (!$c{MONTHS}{$fmonth});
						$filesize{$fname} = $fsize;
						$filedate{$fname} = "$fyear-$c{MONTHS}{$fmonth}-$fday $ftime";
						push (@files, $fname);
					}
				}
			}
			elsif (/^\\/)
			{
				s/\\/\//g;
				$next_root = "/$share$_";
				$next_dir = 1;
			}
		} while ($next_dir == 0 && !eof ($filehandle));
		$path_count{$root}--;

		my $PID = &create_path ($HID, $SID, $root);
		if (defined ($PID))
		{
			my $oldfiles = $db->selectall_arrayref ("SELECT FileName, FileSize, FileDate FROM file WHERE PID=$PID");

			if (@files == 0)
			{
				$db->do ("DELETE FROM file WHERE PID=$PID") if (@$oldfiles > 0);
			}
			elsif (@$oldfiles == 0)
			{
				&insert_files ($PID, \@files, \%filesize, \%filedate);
			}
			else
			{
				my (@insert, @delete, @update);
				my %file_count;
				foreach (@files)
				{
					$file_count{$_}++;
				}
				foreach my $row (@$oldfiles)
				{
					my $name = $row->[0];
					$file_count{$name}--;
					if ($file_count{$name} == 0 && ($row->[1] != $filesize{$name} || $row->[2] ne $filedate{$name}))
					{
						push (@update, $name);
					}
				}
				foreach my $name (keys %file_count)
				{
					if ($file_count{$name} == 1)
					{
						push (@insert, $name);
					}
					elsif ($file_count{$name} == -1)
					{
						push (@delete, $name);
					}
				}
				&purge_files ($PID, \@delete) if (@delete > 0);
				&update_files ($PID, \@update, \%filesize, \%filedate) if (@update > 0);
				&insert_files ($PID, \@insert, \%filesize, \%filedate) if (@insert > 0);
			}
		}
	}

	my @purge_path;
	my $paths = $db->selectall_arrayref ("SELECT PID,FullName FROM path WHERE SID=$SID");
	foreach (@$paths)
	{
		$path_count{$_->[1]}++;
		push (@purge_path, $_->[0]) if ($path_count{$_->[1]} == 1);
	}
	&purge_paths (\@purge_path) if (@purge_path > 0);
	$db->do ("UPDATE status SET UpdatingHost=0");
	$db->do ("UPDATE host SET ExpireCount=0,Counter=0,Flags=Flags|$c{HOST_ONLINE},LastScan=NOW() WHERE HID=$HID");
	&update_share_file_size ($HID, $SID);
	&update_host_file_size ($HID);
	&sem_up ($c{SEM_DB});
}

# scan FTP servers
sub ftp_crawl
{
	my $hosts = &ftp_get_host_list ();
	my $running = 0;
	my ($pid, $host);
	my (%pid2host, %host2pid, %worklist);

	while (@$hosts > 0 || $running > 0)
	{
		$pid = 0;
		undef ($host);
		if ($c{FTP_PARALLEL} == 0 || @$hosts == 0)
		{
			$pid = wait ();
			if ($pid == -1)
			{
				&log_print ("There should be $running child processes but there are none.", 1);
				$c{FTP_PARALLEL} = $running;
				$running = 0;
			}
			else
			{
				$c{FTP_PARALLEL}++;
				$running--;

				$host = $pid2host{$pid};
				delete $pid2host{$pid};
				delete $host2pid{$host};

				if ((scalar @{$worklist{$host}}) > 0)
				{
					$host = shift @{$worklist{$host}};
				}
				else
				{
					delete $worklist{$host};
					undef $host;
				}
			}
		}
		if (!$host && @$hosts > 0)
		{
			$host = shift @$hosts;
		}
		if ($host)
		{
			if ($host2pid{$host->[1]})
			{
				push @{$worklist{$host->[1]}}, [@$host];
			}
			else
			{
				$pid = &ftp_scan_host (@$host);
				if ($pid > 0)
				{
					$c{FTP_PARALLEL}--;
					$running++;
					$pid2host{$pid} = $host->[1];
					$host2pid{$host->[1]} = $pid;
					$worklist{$host->[1]} = [] if (!exists ($worklist{$host->[1]}));
				}
			}
		}
	}
	&log_print ("FTP crawl finished", 1);
}

# get list of FTP servers to scan
sub ftp_get_host_list
{
	&get_host_list_from_master ($c{FTP});
	&delete_or_purge_marked_hosts ($c{FTP});

	my $which = '';
	if (@ARGV > 0)
	{
		foreach (@ARGV)
		{
			my $HIDs = $db->selectcol_arrayref ("SELECT HID FROM ftp WHERE HostName LIKE " . $db->quote ($_));
			foreach (@$HIDs)
			{
				$which .= " OR host.HID=$_";
			}
		}
		$which =~ s/^ OR (.*)/ AND ($1)/;
		if ($which eq '')
		{
			$which = ' AND 0';
		}
	}

	if ($operation == $c{CRAWL_FORCE})
	{
		$db->do ("UPDATE host SET Counter=Period WHERE HostType=$c{FTP} AND Counter<Period $which");
		$db->do ("UPDATE host SET ExpireCount=ExpireCount+1 WHERE HostType=$c{FTP} AND Counter>=Period AND Counter>0 AND ExpireCount<255 $which");
	}
	elsif ($operation == $c{CRAWL_ALL})
	{
		$db->do ("UPDATE host SET Counter=Counter+1 WHERE HostType=$c{FTP} AND Counter<Period $which");
		$db->do ("UPDATE host SET ExpireCount=ExpireCount+1 WHERE HostType=$c{FTP} AND Counter>=Period AND Counter>0 AND ExpireCount<255 $which");
	}
	my $unset_flags = $c{HOST_DELETE} | $c{HOST_PURGE};
	my $hosts = $db->selectall_arrayref ("SELECT host.HID,HostName,Login,PassWord,Port,StartingDir,FileList,Flags FROM ftp LEFT JOIN host USING (HID) WHERE ExpireCount>0 AND Counter>=Period AND Counter>0 AND (Flags&$unset_flags)=0 $which ORDER BY ScanOrder,HostName,Login,StartingDir");
	return $hosts;
}

# scan FTP server
# \param $HID Host ID of the server
# \param $host Host name of the server
# \param $login Login to access the server
# \param $password Password to login to the server
# \param $post Port of the FTP server
# \param $starting_dir Directory to start scanning in.
# \param $file_list Name of the file containing the file list.
# \param $flags Flags determinig the scan method
sub ftp_scan_host
{
	my ($HID, $host, $login, $password, $port, $starting_dir, $file_list, $flags) = @_;

	my $pid = fork ();
	if (!defined ($pid))
	{
		&log_print ("Could not fork, reason: $!", 0);
		return 0;
	}
	if ($pid == 0)
	{
		&db_connect (\%c, \$mdb, \$db, \&log_print);
		exit 1 if (!$db);

		my $temp = &get_temp_file_name ();
		my $parser = 0;
		my $ip = &ftp_get_ip ($host);
		my $error = 1 if (!defined ($ip));

		if (!$error)
		{
			&get_time ();
			$error = &ftp_get_listing ($host, $ip, $login, $password, $port,
					$starting_dir, $file_list, $temp,
					$flags & $c{HOST_PASSIVE_MODE},
					$flags & $c{HOST_LS_RECURSIVE});
		}

		if (!$error)
		{
			my $delete = 0;
			my $parser = \&ftp_parser_ffsearch;
			my $basedir = '';

			open (IN, "<$temp");
			if (eof (IN))
			{
				&log_print ("$host: file listing is empty.", 1);
				$delete = 1;
			}
			close (IN);

			if ($file_list)
			{
				$parser = &ftp_detect_parser ($temp) if (!$delete);
				if (!$parser)
				{
					&log_print ("$host:$file_list: unknown format of file listing.", 1);
					$delete = 1;
				}
				$basedir = $file_list;
				$basedir =~ s/\/*[^\/]+$//;
			}
			elsif ($ftp_ls_recursive)	# $ftp_ls_recursive is set in ftp_get_listing
			{
				$parser = &ftp_detect_parser ($temp) if (!$delete);
				if (!$parser)
				{
					&log_print ("$host: unknown format of recursive file listing.", 1);
					$delete = 1;
				}
				$basedir = $starting_dir;
			}

			if ($delete)
			{
				&sem_down ($c{SEM_DB});
				$db->do ("UPDATE host SET IP='$ip' WHERE HID=$HID");
				$mdb->do ("UPDATE host SET IP='$ip' WHERE HID=$HID") if ($mdb != $db);
				&delete_host ($HID);
				&sem_up ($c{SEM_DB});
			}
			else
			{
				open (IN, "<$temp");
				&sem_down ($c{SEM_DB});
				&log_print ("$host: updating database", 1);
				$db->do ("UPDATE status SET UpdatingHost=$HID");
				&$parser ($HID, $basedir, *IN);
				$db->do ("UPDATE status SET UpdatingHost=0");
				$db->do ("UPDATE host SET IP='$ip',ExpireCount=0,Counter=0,Flags=Flags|$c{HOST_ONLINE},LastScan=NOW() WHERE HID=$HID");
				$mdb->do ("UPDATE host SET IP='$ip' WHERE HID=$HID") if ($mdb != $db);
				&update_host_file_size ($HID);
				&sem_up ($c{SEM_DB});
				close (IN);
			}
		}
		else
		{
			&sem_down ($c{SEM_DB});
			$db->do ("UPDATE host SET IP='$ip' WHERE HID=$HID");
			$mdb->do ("UPDATE host SET IP='$ip' WHERE HID=$HID") if ($mdb != $db);
			if ($error == 2)
			{
				$db->do ("UPDATE host SET Counter=Period-1 WHERE HID=$HID AND Period>0");
				$db->do ("UPDATE host SET Flags=Flags|$c{HOST_ONLINE},LastScan=NOW() WHERE HID=$HID");
			}
			my $expire = $db->selectrow_array ("SELECT ExpireCount FROM host WHERE HID=$HID");
			if ($expire >= $c{FTP_EXPIRE})
			{
				&log_print ("$host: expired, deleting.", 1);
				&delete_host ($HID);
			}
			&sem_up ($c{SEM_DB});
		}

		&db_disconnect ($mdb, $db);
		unlink ($temp);
		exit 0;
	}
	return $pid;
}

# get a list of files from FTP server
# \param $host Host name of the FTP server
# \param $ip IP address of the FTP server
# \param $login Login to access the server
# \param $password Password to login to the server
# \param $post Port of the FTP server
# \param $starting_dir Directory to start scanning in.
# \param $file_list Name of the file containing the file list.
# \param $temp Name of the temporary file
# \param $passive Use passive mode to scan the server
# \param $ls_recursive Use recursive ls to scan the server
sub ftp_get_listing
{
	my ($host, $ip, $login, $password, $port, $starting_dir, $file_list, $temp, $passive, $ls_recursive) = @_;
	my $error = 0;

	&log_print ("$host: crawling ($ip:$port, $login, $password)", 1);
	$ftp_host = $host;
	$ftp_ip = $ip;
	$ftp_login = $login;
	$ftp_password = $password;
	$ftp_port = $port;
	$ftp_passive = $passive;
	$ftp_ls_recursive = $ls_recursive;
	$error = &ftp_connect ();
	return $error if ($error);

	if ($file_list)
	{
		my $suffix = '';
		$suffix = '.gz' if ($file_list =~ /\.gz$/);
		$suffix = '.bz2' if ($file_list =~ /\.bz2$/);
		$suffix = '.Z' if ($file_list =~ /\.Z$/);

		$ftp_handle->binary ();
		$error = 1 if (!$ftp_handle->get ($file_list, $temp . $suffix));
		if ($error)
		{
			&log_print ("$host:$file_list: can't read file listing.", 1);
		}
		else
		{
			system ("gzip -f -d $temp$suffix") if ($suffix eq '.gz');
			system ("bzip2 -f -d $temp$suffix") if ($suffix eq '.bz2');
			system ("uncompress -f $temp$suffix") if ($suffix eq '.Z');
		}
	}
	else
	{
		$error = &ftp_chdir ($ftp_handle, $starting_dir);
		if (!$error)
		{
			my $out;
			my $out_short;
			if ($ftp_ls_recursive)
			{
				my $reconnections;
				my $max_reconnections = 1;

				$reconnections = 0;
				while (1)
				{
					$out_short = $ftp_handle->dir ();
					if (!&ftp_connection_closed ())
					{
						last;
					} else {
						return 1 if ($reconnections >= $max_reconnections);
						$reconnections++;
						$ftp_handle->close ();
						$error = &ftp_connect ();
						return $error if ($error);
						$error = &ftp_chdir ($ftp_handle, $starting_dir);
						return $error if ($error);
					}
				}
				$reconnections = 0;
				while (1)
				{
					$out = $ftp_handle->dir ("-lR");
					if (!&ftp_connection_closed ())
					{
						last;
					} else {
						return 1 if ($reconnections >= $max_reconnections);
						$reconnections++;
						$ftp_handle->close ();
						$error = &ftp_connect ();
						return $error if ($error);
						$error = &ftp_chdir ($ftp_handle, $starting_dir);
						return $error if ($error);
					}
				}

				if (!$out_short || !$out
				    || 2 * @$out_short + 2 >= @$out)
				{
					# output of "ls -lR" is not significantly
					# longer than output of "ls" so do not
					# use "ls -lR"
					$ftp_ls_recursive = 0;
				}
			}

			$file_listing_size = 0;
			open (OUT, ">$temp");

			if ($ftp_ls_recursive)
			{
				my $cur_dir = $ftp_handle->pwd ();
				$cur_dir =~ s/\/+$//;
				$cur_dir = "." if (!$cur_dir);
				if ($$out[0] eq '')
				{
					shift @$out;
				}
				if ($$out[0] !~ /:$/)
				{
					print OUT "$cur_dir:\n";
				}
				foreach (@$out)
				{
					print OUT "$_\n";
				}
			}
			else
			{
				$error = &ftp_scan_dir (0, '', $starting_dir);
			}
			close (OUT);
		}
	}

	$ftp_handle->close () if (defined $ftp_handle);
	return $error;
}

# connect to FTP server
sub ftp_connect
{
	$ftp_handle = Net::FTP->new ($ftp_ip, Port => $ftp_port, Timeout => $c{FTP_TIMEOUT}, Passive => $ftp_passive);
	if (!defined $ftp_handle)
	{
		&log_print ("$ftp_host: unreachable, reason: $@", 1);
		return 1;
	}

	# get the server identification
	my $message = $ftp_handle->message ();
	if ($message =~ /PureFTPd/i || $message =~ /Pure-FTPd/i	# PureFTPd sends only 2000 lines of "ls -lR"
	    || $message =~ /CesarFTP/i		# CesarFTP does not print "<DIRECTORY>:"
	    || $message =~ /GuildFTPd/i		# GuildFTPd gets crazy when attempting to use "ls -lR"
	    || $message =~ /Microsoft/i		# FFSearch can't parse output of "ls -lR" of Microsoft FTP server
	    || $message =~ /Serv-U/i		# Serv-U's "ls -lR" does not work well in all cases
	    )
	{
		$ftp_ls_recursive = 0;
	}

	if ($ftp_password eq 'user@host.com')
	{
		$ftp_password = $c{FTP_ANON_PASSWD};
	}
	if ($ftp_password =~ /^['"](.*)['"]$/)
	{
		$ftp_password = $1;
	}
	if (!$ftp_handle->login ($ftp_login, $ftp_password))
	{
		&log_print ("$ftp_host:$ftp_port: can't login: $ftp_login:$ftp_password", 1);
		$ftp_handle->close ();
		return 2;
	}
	$ftp_cur_dir = "";
	return 0;
}

# return true if the connection to FTP server was closed
sub ftp_connection_closed
{
	my $code = $ftp_handle->code ();

	&log_print ("$ftp_host:$ftp_port: timeout", 1) if ($code eq '000');

	return ($code eq '599' || $code eq '000');
}

# get IP address of FTP server
# \param $host Name of the server
sub ftp_get_ip
{
	my ($host) = @_;

	return $host if (!&is_invalid_IP ($host));

	my @addr = (gethostbyname ($host))[4];
	if (!@addr)
	{
		&log_print ("$host: can't get IP address\n", 1);
		return undef;
	}
	return join ('.', unpack ('C4', $addr[0]));
}

# Check whether the dir was changed.
sub ftp_chdir_succeeded
{
	my $new_dir = $ftp_handle->pwd ();
	my $ret = $new_dir ne $ftp_cur_dir;
	$ftp_cur_dir = $new_dir;

	return $ret;
}

# change to path by the parts
# \param $ftp_handle Handle connected to FTP server
# \param $path Path to enter to
sub ftp_chdir
{
	my ($ftp_handle, $path) = @_;

	my $dir = '';
	$ftp_cur_dir = $ftp_handle->pwd ();

	my @parts = split (/\/+/, $path);
	foreach (@parts)
	{
		next if (!$_);
		$dir .= '/' . $_;
		if (!$ftp_handle->cwd ($_) || !&ftp_chdir_succeeded ())
		{
			&log_print ("Could not chdir to $dir", 0);
			return 1;
		}
	}

	return 0;
}

# scan directory of FTP server
# \param $depth Depth of the directory
# \param $shortname Short name of the directory
# \param $fullname Full name (complete path) of the directory
sub ftp_scan_dir
{
	my ($depth, $shortname, $fullname) = @_;

	&log_print ("ftp_scan_dir: scanning '$fullname'", 2);

	if ($depth > $c{FTP_MAX_DIR_DEPTH})
	{
		&log_print ("$ftp_host:$ftp_port:$fullname: directory tree too deep, there may be a directory loop.", 0);
		return 1;
	}

	my $error = 0;
	my $upper_dir = $fullname;
	$upper_dir =~ s/\/[^\/]*$//;

	my $reconnections = 0;
	my $max_reconnections = 1;	# max reconnections in current dir

ftp_scan_dir_again:
	if ($shortname ne '')
	{
		if (!$ftp_handle->cwd ($shortname) || !&ftp_chdir_succeeded ())
		{
			if (&ftp_connection_closed ())
			{
				return 1 if ($reconnections >= $max_reconnections);
				$reconnections++;
				$ftp_handle->close ();
				$error = &ftp_connect ();
				return $error if ($error);
				$error = &ftp_chdir ($ftp_handle, $upper_dir);
				return $error if ($error);
				goto ftp_scan_dir_again;
			}

			&log_print ("Could not chdir to $fullname", 1);
			return 0;
		}
	}

	my $out = $ftp_handle->dir ();
	if (&ftp_connection_closed ())
	{
		return 1 if ($reconnections >= $max_reconnections);
		$reconnections++;
		$ftp_handle->close ();
		$error = &ftp_connect ();
		return $error if ($error);
		$error = &ftp_chdir ($ftp_handle, $upper_dir);
		return $error if ($error);
		goto ftp_scan_dir_again;
	}
	my @scan;

	print OUT "$fullname\n";
	$file_listing_size += length ($fullname) + 1;
	if ($c{MAX_TEMP_FILE_SIZE} && $file_listing_size > $c{MAX_TEMP_FILE_SIZE})
	{
		&log_print ("$ftp_host:$ftp_port: file listing is too long, there may be a directory loop.", 0);
		return 1;
	}

	foreach (@$out)
	{
		if (/^([d\-])\S{9}\s+\S+\s+\S+\s+\S+\s+(\d+)\s+(.{12}) (.*?)\r?$/		# UNIX format
			|| /^([d\-])\S{9}\s+\S+\s+\S+\s+(\d+)\s+(.{12}) (.*?)\r?$/		# similar to UNIX format
			|| /^([d\-])\s+\[[^\]]+\]\s+\S+\s+(\d+)\s+(.{12}) (.*?)\r?$/		# Novell Netware format
			|| /^([d\-])\[[^\]]+\]\s+\d+\s+\S+\s+(\d+)\s+(.{12}) (.*?)\r?$/)	# another Novell Netware format
		{
			my $name = $4;
			next if (($name eq ".") || ($name eq "..") || ($name eq ""));
			my $ftype = $1;
			if ($ftype eq '-')
			{
				my $fsize = $2;
				my $fdatetime = &ftp_parse_unix_datetime ($3);
				$_ = "\t$fdatetime $fsize $name\n";
			}
			elsif ($ftype eq 'd')
			{
				push (@scan, $name);
				next;
			}
		}
		elsif (/^([d\-])\S{9}\s+\S+\s+\S+\s+\S+\s+(\d+)\s+(\d{4}-\d\d-\d\d \d\d:\d\d) ([^\/]*?)\/?\r?$/)    # similar to UNIX format
		{
			my $name = $4;
			next if (($name eq ".") || ($name eq "..") || ($name eq ""));
			my $ftype = $1;
			if ($ftype eq '-')
			{
				my $fsize = $2;
				my $fdatetime = $3 . ":00";
				$_ = "\t$fdatetime $fsize $name\n";
			}
			elsif ($ftype eq 'd')
			{
				push (@scan, $name);
				next;
			}
		}
		elsif (/^(\d\d)-(\d\d)-(\d\d)\s+(\d\d):(\d\d)([AP])M\s+(\d+) (.*?)\r?$/)	# Microsoft format - files
		{
			my $name = $8;
			next if ($name eq "");
			my $fmonth = $1;
			my $fday = $2;
			my $fyear = $3;
			my $fhour = $4;
			my $fmin = $5;
			my $fdaypart = $6;
			my $fsize = $7;
			$fyear += ($fyear < 80) ? 2000 : 1900;
			$fhour += 12 if ($fdaypart eq 'P');
			$_ = "\t$fyear-$fmonth-$fday $fhour:$fmin:00 $fsize $name\n";
		}
		elsif (/^\d\d-\d\d-\d\d\s+\d\d:\d\d[AP]M\s+[<]DIR[>]\s+(.*?)\r?$/)	# Microsoft format - directories
		{
			my $name = $1;
			next if (($name eq ".") || ($name eq "..") || ($name eq ""));
			push (@scan, $name);
			next;
		}
		else
		{
			next;
		}

		# print the info about file
		print OUT $_;
		$file_listing_size += length ($_);
		if ($c{MAX_TEMP_FILE_SIZE} && $file_listing_size > $c{MAX_TEMP_FILE_SIZE})
		{
			&log_print ("$ftp_host: file listing is too long, there may be a directory loop.", 1);
			return 1;
		}
	}

	$error = 0;
	foreach (@scan)
	{
		$error = &ftp_scan_dir ($depth + 1, $_, "$fullname/$_");
		return $error if ($error);
	}

	if ($shortname ne '')
	{
		if (!$ftp_handle->cdup () || !&ftp_chdir_succeeded ())
		{
#			if (&ftp_connection_closed ())
#			{
				return 1 if ($reconnections >= $max_reconnections);
				$reconnections++;
				$ftp_handle->close ();
				$error = &ftp_connect ();
				return $error if ($error);
				return &ftp_chdir ($ftp_handle, $upper_dir);
#			}

			&log_print ("Could not chdir to $fullname/..", 1);
			return 1;
		}
	}
	return 0;
}

# parse date & time and return it in format applicable to database
# \param $datetime date & time to parse
sub ftp_parse_unix_datetime
{
	my ($datetime) = @_;

	my ($month, $day, $time);
	my ($year, $hour, $min);

	if ($datetime =~ /^(\S+)\s+(\d+)\s+(\S+)$/)
	{
		$month = $1;
		$day = $2;
		$time = $3;
	}
	else
	{
		return "0000-00-00 00:00:00";
	}

# When date is older than 6 months or in future
# the format is "%b %d  %Y" (for example "Dec 24  2003", see man strftime),
# otherwise the format is "%b %d %R" (for example "Jul 24 12:34")

	$month = $c{MONTHS}{$month};
	$month = "00" if (!$month);
	if ($time =~ /^(\d+):(\d+)$/)
	{
		$hour = $1;
		$min = $2;

		$hour = "0$hour" if (length ($hour) == 1);
		$min = "0$min" if (length ($min) == 1);
		$time = "$hour:$min:00";

		# If the month is more than 1 month greater it is date from
		# near past (specification says 6 months).  Using 11 months
		# is OK as server should have selected correct format.
		# This also avoids bad parsing if the date is from different
		# time zone, for example in the following case
		# local date & time:   Jan 31, 23:23
		# remote date & time:  Feb 1, 1:23
		# simple ($month > $c{month}) does not work correctly.
		if ($month > $c{month} + 1)
		{
			$year = $c{year} - 1;
		}
		else
		{
			$year = $c{year};
		}
	}
	elsif ($time =~ /^\d+$/)
	{
		$year =  $time;
		$time = '00:00:00';
	}
	else
	{
		$year = $c{year};
		$time = '00:00:00';
	}
	$day = "0$day" if (length ($day) == 1);
	return "$year-$month-$day $time";
}

# detect parser of FTP server output
# \param $temp Name of temporarty file containing the file listing
sub ftp_detect_parser
{
	my ($temp) = @_;

	my $line;
	my $count;

	# detect UNIX "ls -lR" format and similar formats
	open (IN , "<$temp");
	chomp ($line = <IN>);
	if ($line =~ /:$/
		|| $line =~ /^total \d+$/i
		|| $line =~ /^\S{10}\s+\S+\s+\S+\s+\S+\s+(\d+)\s+(.{12}) (.*?)\r?$/				# UNIX "ls -lR" format
		|| $line =~ /^\S{10}\s+\S+\s+\S+\s+(\d+)\s+(.{12}) (.*?)\r?$/					# similar to UNIX format
		|| $line =~ /^\S{10}\s+\S+\s+\S+\s+\S+\s+(\d+)\s+(\d{4}-\d\d-\d\d \d\d:\d\d) (.*?)\r?$/)	# similar to UNIX format
	{
		close (IN);
		return \&ftp_parser_unix_ls_lR;
	}
	close (IN);

	return undef;
}

# parser for FFSearch's internal format
# processes the output and updates the database
# \param $HID Host ID of the FTP server
# \param $basedir Directory that the crawling started in
# \param $filehandle Handle of the file containing file listing
sub ftp_parser_ffsearch
{
	my ($HID, $basedir, $filehandle) = @_;

	my %path_count;
	my (@files, %filedate, %filesize);
	my $next_path = <$filehandle>;
	my $next;
	chomp ($next_path);
	while (!eof ($filehandle))
	{
		my $path = $next_path;
		$next = 0;

		@files = ();
		%filedate = ();
		%filesize = ();
		do {
			$_ = <$filehandle>;
			chomp;
			if (/^\t(\d{4}-\d\d-\d\d \d\d:\d\d:\d\d) (\d+) (.+)$/)
			{
				my $fname = $3;
				$filesize{$fname} = $2;
				$filedate{$fname} = $1;
				push (@files, $fname);
			}
			elsif (!/^\t/)
			{
				$next_path = $_;
				$next = 1;
			}
		} while ($next == 0 && !eof ($filehandle));
		$path_count{$path}--;
		&ftp_update_database ($HID, $path, \@files, \%filesize, \%filedate);
	}
	if ($next)
	{
		@files = ();
		%filedate = ();
		%filesize = ();
		$path_count{$next_path}--;
		&ftp_update_database ($HID, $next_path, \@files, \%filesize, \%filedate);
	}
	&ftp_purge_paths ($HID, \%path_count);
}

# parser for UNIX "ls -lR" format
# processes the output and updates the database
# \param $HID Host ID of the FTP server
# \param $basedir Directory that the crawling started in
# \param $filehandle Handle of the file containing file listing
sub ftp_parser_unix_ls_lR
{
	my ($HID, $basedir, $filehandle) = @_;

	my $skip_read = 0;
	my %path_count;
	my (@files, %filedate, %filesize);
	my $next_path;
	my $next;

	$next_path = <$filehandle>;
	chomp $next_path;
	if ($next_path =~ /:$/)
	{
		$next_path =~ s/\/*:$//;
	}
	elsif ($next_path =~ /^total \d+$/i)
	{
		$next_path = ".";
	}
	else
	{
		$next_path = ".";
		$skip_read = 1;
	}

	my $orig_basedir = quotemeta ($next_path);
	while (!eof ($filehandle))
	{
		my $path = $next_path;
		$next = 0;
		$path = "/" . $path if ($path !~ /^\/|^\.$|^\.[^\.]/);
		$path =~ s/^$orig_basedir/$basedir/;
		$path =~ s/\/+$//;

		@files = ();
		%filedate = ();
		%filesize = ();
		do {
			if ($skip_read == 0)
			{
				$_ = <$filehandle>;
				chomp;
			}
			$skip_read = 0;

			if (/^-\S{9}\s+\S+\s+\S+\s+\S+\s+(\d+)\s+(.{12}) (.*?)\r*$/
				|| /^-\S{9}\s+\S+\s+\S+\s+(\d+)\s+(.{12}) (.*?)\r?$/)	# similar to UNIX format
			{
				my $name = $3;
				next if (($name eq ".") || ($name eq "..") || ($name eq ""));
				$filesize{$name} = $1;
				$filedate{$name} = &ftp_parse_unix_datetime ($2);
				push (@files, $name);
			}
			elsif (/^-\S{9}\s+\S+\s+\S+\s+\S+\s+(\d+)\s+(\d{4}-\d\d-\d\d \d\d:\d\d) ([^\/]*?)\r?$/)    # similar to UNIX format
			{
				my $name = $3;
				next if (($name eq ".") || ($name eq "..") || ($name eq ""));
				$filesize{$name} = $1;
				$filedate{$name} = $2 . ":00";
				push (@files, $name);
			}
			elsif (/(.+):$/)
			{
				$next_path = $1;
				$next = 1;
			}
		} while ($next == 0 && !eof ($filehandle));
		$path_count{$path}--;
		&ftp_update_database ($HID, $path, \@files, \%filesize, \%filedate);
	}
	if ($next)
	{
		@files = ();
		%filedate = ();
		%filesize = ();
		$path_count{$next_path}--;
		&ftp_update_database ($HID, $next_path, \@files, \%filesize, \%filedate);
	}
	&ftp_purge_paths ($HID, \%path_count);
}

# delete paths that exist in database but not on FTP server
# \param $HID Host ID of the FTP server
# \param $path_count Associative array containing each path located on the server
sub ftp_purge_paths
{
	my ($HID, $path_count) = @_;

	my @purge_path;
	my $paths = $db->selectall_arrayref ("SELECT PID,FullName FROM path WHERE HID=$HID");
	foreach (@$paths)
	{
		$path_count->{$_->[1]}++;
		push (@purge_path, $_->[0]) if ($path_count->{$_->[1]} == 1);
	}
	&purge_paths (\@purge_path) if (@purge_path > 0);
}

# update contents of FTP server in database
# \param $HID Host ID of the server
# \param $path Complete path on the server
# \param $files List of files in the directory
# \param $filesize Sizes of the files
# \param $filedate Dates & times od the files
sub ftp_update_database
{
	my ($HID, $path, $files, $filesize, $filedate) = @_;

	my $PID = &create_path ($HID, 0, $path);
	if (defined ($PID))
	{
		my $oldfiles = $db->selectall_arrayref ("SELECT FileName, FileSize, FileDate FROM file WHERE PID=$PID");

		if (@$files == 0)
		{
			$db->do ("DELETE FROM file WHERE PID=$PID") if (@$oldfiles > 0);
		}
		elsif (@$oldfiles == 0)
		{
			&insert_files ($PID, $files, $filesize, $filedate);
		}
		else
		{
			my (@insert, @delete, @update);
			my %file_count;

			foreach (@$files)
			{
				$file_count{$_}++;
			}
			foreach my $row (@$oldfiles)
			{
				my $name = $row->[0];
				$file_count{$name}--;
				if ($file_count{$name} == 0)
				{
					if ($row->[1] != $filesize->{$name} || $row->[2] lt $filedate->{$name})
					{
						push (@update, $name);
					}
					elsif ($row->[2] gt $filedate->{$name})
					{
						$db->do ("UPDATE file SET FileDate=". $db->quote ($filedate->{$name}) . " WHERE FileName=" . $db->quote ($name) . " AND PID=$PID");
					}
				}
			}
			foreach my $name (keys %file_count)
			{
				if ($file_count{$name} == 1)
				{
					push (@insert, $name);
				}
				elsif ($file_count{$name} == -1)
				{
					push (@delete, $name);
				}
			}
			&purge_files ($PID, \@delete) if (@delete > 0);
			&update_files ($PID, \@update, $filesize, $filedate) if (@update > 0);
			&insert_files ($PID, \@insert, $filesize, $filedate) if (@insert > 0);
		}
	}
}

# add path to database
# \param $HID Host ID of the server
# \param $SID ID of the share on the  server
# \param $path Path on the server
sub create_path
{
	my ($HID, $SID, $path) = @_;

	# good hash function is needed here, CRC32 is excelent for this purpose
	my $hash = crc32 ($path, crc32 (pack ("II", $HID, $SID)));
	my $qpath = $db->quote ($path);
	my $select = "SELECT PID FROM path WHERE HID=$HID AND SID=$SID AND `Hash`=$hash AND FullName=$qpath";
	my $PID = $db->selectrow_array ($select);
	if (!defined ($PID))
	{
		my $name = $path;
		my $parent = $path;
		my $PSID = $SID;
		$name =~ s/.*\///;
		$parent =~ s/\/*[^\/]*$//;
		$PSID = 0 if ($parent eq '');
		my $phash = crc32 ($parent, crc32 (pack ("II", $HID, $PSID)));
		my $PPID = $db->selectrow_array ("SELECT PID FROM path WHERE HID=$HID AND SID=$PSID AND `Hash`=$phash AND FullName=" . $db->quote ($parent));
		$PPID = 0 if (!defined ($PPID));
		$db->do ("INSERT INTO path (`Hash`,PPID,HID,SID,ShortName,FullName) VALUES ($hash,$PPID,$HID,$SID," . $db->quote ($name) . ",$qpath)");
		$PID = $db->selectrow_array ($select);
	}
	return $PID;
}

# get list of servers from master database
# \param $type Type of the server (SMB or FTP)
sub get_host_list_from_master
{
	my ($type) = @_;

	if ($mdb != $db)
	{
		my ($mhosts, $hosts, $row, $global_flags, $local_flags);
		if ($type == $c{SMB})
		{
			$global_flags = $c{HOST_DELETE} | $c{HOST_PURGE} | $c{HOST_MANUAL_CHANGES};
			$local_flags = $c{HOST_ONLINE};
			$mhosts = $mdb->selectall_hashref ("SELECT host.HID,HostName,WorkGroup,IP,Period,ScanOrder,Flags FROM smb LEFT JOIN host USING (HID)",'HID');
			foreach (keys %$mhosts)
			{
				$row = $mhosts->{$_};
				$row->{Flags} &= $global_flags;
				$db->do ("LOCK TABLES host WRITE, ftp WRITE, smb WRITE");
				my ($hosttype, $hostname) = $db->selectrow_array ("SELECT HostType,HostName FROM host LEFT JOIN smb USING (HID) WHERE host.HID=$_");
				if ($hosttype != $type)
				{
					$db->do ("DELETE FROM smb WHERE HID=$_");
					$db->do ("DELETE FROM ftp WHERE HID=$_");
					$db->do ("REPLACE INTO host (HID,IP,HostType,ExpireCount,Period,Counter,ScanOrder,Flags) VALUES ($_," . $db->quote ($row->{IP}) . ",$c{SMB},1,$row->{Period},$row->{Period},$row->{ScanOrder},$row->{Flags})");
					$db->do ("INSERT INTO smb (HID,HostName,WorkGroup) VALUES ($_," . $db->quote ($row->{HostName}) . "," . $db->quote ($row->{WorkGroup}) . ")");
					$db->do ("UNLOCK TABLES");
					&delete_host ($_);
				}
				else
				{
					$db->do ("UPDATE host SET IP=" . $db->quote ($row->{IP}) . ",Period=$row->{Period},ScanOrder=$row->{ScanOrder},Flags=(Flags&$local_flags)|$row->{Flags} WHERE HID=$_");
					$db->do ("UPDATE smb SET HostName=" . $db->quote ($row->{HostName}) . ",WorkGroup=" . $db->quote ($row->{WorkGroup}) . " WHERE HID=$_");
					$db->do ("UNLOCK TABLES");
					&delete_host ($_) if ($row->{HostName} ne $hostname);
				}
			}
			$hosts = $db->selectcol_arrayref ("SELECT HID FROM host WHERE HostType=$type");
			foreach (@$hosts)
			{
				if (!$mhosts->{$_})
				{
					&purge_host ($_);
				}
			}
		}
		elsif ($type == $c{FTP})
		{
			$global_flags = $c{HOST_DELETE} | $c{HOST_PURGE} | $c{HOST_PASSIVE_MODE} | $c{HOST_LS_RECURSIVE};
			$local_flags = $c{HOST_ONLINE};
			$mhosts = $mdb->selectall_hashref ("SELECT host.HID,HostName,Login,PassWord,Port,StartingDir,FileList,Comment,EditableBy,IP,Period,ScanOrder,Flags FROM ftp LEFT JOIN host USING (HID)",'HID');
			foreach (keys %$mhosts)
			{
				$row = $mhosts->{$_};
				$row->{Flags} &= $global_flags;
				$db->do ("LOCK TABLES host WRITE, ftp WRITE, smb WRITE");
				my ($hosttype, $hostname, $login, $port, $starting_dir) = $db->selectrow_array ("SELECT HostType,HostName,Login,Port,StartingDir FROM host LEFT JOIN ftp USING (HID) WHERE host.HID=$_");
				if ($hosttype != $type)
				{
					$db->do ("DELETE FROM smb WHERE HID=$_");
					$db->do ("DELETE FROM ftp WHERE HID=$_");
					$db->do ("REPLACE INTO host (HID,IP,HostType,ExpireCount,Period,Counter,ScanOrder,Flags) VALUES ($_," . $db->quote ($row->{IP}) . ",$c{FTP},1,$row->{Period},$row->{Period},$row->{ScanOrder},$row->{Flags})");
					$db->do ("INSERT INTO ftp (HID,HostName,Login,PassWord,Port,StartingDir,FileList,Comment,EditableBy) VALUES ($_," . $db->quote ($row->{HostName}) . "," . $db->quote ($row->{Login}) . "," . $db->quote ($row->{PassWord}) . ",$row->{Port}," . $db->quote ($row->{StartingDir}) . "," . $db->quote ($row->{FileList}) . "," . $db->quote ($row->{Comment}) . "," . $db->quote ($row->{EditableBy}) . ")");
					$db->do ("UNLOCK TABLES");
					&delete_host ($_);
				}
				else
				{
					$db->do ("UPDATE host SET IP=" . $db->quote ($row->{IP}) . ",Period=$row->{Period},ScanOrder=$row->{ScanOrder},Flags=(Flags&$local_flags)|$row->{Flags} WHERE HID=$_");
					$db->do ("UPDATE ftp SET HostName=" . $db->quote ($row->{HostName}) . ",Login=" . $db->quote ($row->{Login}) . ",PassWord=" . $db->quote ($row->{PassWord}) . ",Port=$row->{Port},StartingDir=" . $db->quote ($row->{StartingDir}) . ",FileList=" . $db->quote ($row->{FileList}) . ",Comment=" . $db->quote ($row->{Comment}) . ",EditableBy=" . $db->quote ($row->{EditableBy}) . " WHERE HID=$_");
					$db->do ("UNLOCK TABLES");
					&delete_host ($_) if ($row->{HostName} ne $hostname || $row->{Login} ne $login || $row->{Port} != $port || $row->{StartingDir} != $starting_dir);
				}
			}
			$hosts = $db->selectcol_arrayref ("SELECT HID FROM host WHERE HostType=$type");
			foreach (@$hosts)
			{
				if (!$mhosts->{$_})
				{
					&purge_host ($_);
				}
			}
		}
	}
}

# delete or purge marked hosts
# \param $type Type of the server (SMB or FTP)
sub delete_or_purge_marked_hosts
{
	my ($type) = @_;

	&log_print ("Deleting hosts that are scheduled for deleting.", 1);
	my $flags_wanted = $c{HOST_DELETE} | $c{HOST_PURGE};
	my $hosts = $db->selectall_arrayref ("SELECT HID,Flags FROM host WHERE HostType=$type AND (Flags&$flags_wanted)");
	foreach (@$hosts)
	{
		if (($_->[1]) & $c{HOST_PURGE})
		{
			&purge_host ($_->[0]);
		}
		else
		{
			&delete_host ($_->[0]);
		}
	}
}

# purge (completely delete) server from database
# \param $HID Host ID of the server
sub purge_host
{
	my ($HID) = @_;

	my $PIDs = $db->selectcol_arrayref ("SELECT PID FROM path WHERE HID=$HID");
	foreach (@$PIDs)
	{
		$db->do ("DELETE FROM file WHERE PID=$_");
	}
	$db->do ("DELETE FROM path WHERE HID=$HID");
	$db->do ("DELETE FROM share WHERE HID=$HID");
	$db->do ("DELETE FROM smb WHERE HID=$HID");
	$db->do ("DELETE FROM ftp WHERE HID=$HID");
	$db->do ("DELETE FROM host WHERE HID=$HID");
}

# delete contents of the server from database
# \param $HID Host ID of the server
sub delete_host
{
	my ($HID) = @_;

	my $PIDs = $db->selectcol_arrayref ("SELECT PID FROM path WHERE HID=$HID");
	foreach (@$PIDs)
	{
		$db->do ("DELETE FROM file WHERE PID=$_");
	}
	$db->do ("DELETE FROM path WHERE HID=$HID");
	$db->do ("DELETE FROM share WHERE HID=$HID");
	$db->do ("UPDATE host SET Flags=Flags&~($c{HOST_ONLINE}),LastScan='',TotalFileSize=-1 WHERE HID=$HID");
}

# purge shares from database
# \param $SIDs Reference to list of Share IDs
sub purge_shares
{
	my ($SIDs) = @_;
	foreach my $SID (@$SIDs)
	{
		if ($SID != 0)
		{
			my $PIDs = $db->selectcol_arrayref ("SELECT PID FROM path WHERE SID=$SID");
			foreach (@$PIDs)
			{
				$db->do ("DELETE FROM file WHERE PID=$_");
			}
			$db->do ("DELETE FROM path WHERE SID=$SID");
			$db->do ("DELETE FROM share WHERE SID=$SID");
		}
	}
}

# purge paths from database
# \param $PIDs Reference to list of Path IDs
sub purge_paths
{
	my ($PIDs) = @_;

	foreach (@$PIDs)
	{
		$db->do ("DELETE FROM file WHERE PID=$_");
		$db->do ("DELETE FROM path WHERE PID=$_");
	}
}

# purge files from database
# \param $PID ID of the path of directory containing the files
# \param $files Reference to list of file names
sub purge_files
{
	my ($PID, $files) = @_;

	my $exec;
	my $count = 0;
	foreach (@$files)
	{
		$exec .= ($count++ > 0) ? ' OR FileName=' . $db->quote ($_)
			: "DELETE FROM file WHERE PID=$PID AND (FileName=" . $db->quote ($_);
		if ($count >= $c{DB_BATCH})
		{
			$db->do ($exec.')');
			$exec = '';
			$count = 0;
		}
	}
	$db->do ($exec.')') if ($count);
}

# insert files to database
# \param $PID ID of the path of directory containing the files
# \param $files Reference to list of file names
# \param $filesize Sizes of the files
# \param $filedate Dates & times od the files
sub insert_files
{
	my ($PID, $files, $filesize, $filedate) = @_;

	my $exec;
	my $count = 0;
	foreach (@$files)
	{
		my $end = substr ($_, -4, 4);
		$exec .= ($count++ > 0) ? ',' : 'INSERT INTO file (PID,FileName,FileNameEnd,DateAdded,FileSize,FileDate) VALUES ';
		$exec .= "($PID," . $db->quote ($_) . ",REVERSE(" . $db->quote ($end) . "),UNIX_TIMESTAMP(),$filesize->{$_}," . $db->quote ($filedate->{$_}) . ")";
		if ($count >= $c{DB_BATCH})
		{
			$db->do ($exec);
			$exec = '';
			$count = 0;
		}
	}
	$db->do ($exec) if ($count);
}

# update files in database
# \param $PID ID of the path of directory containing the files
# \param $files Reference to list of file names
# \param $filesize Sizes of the files
# \param $filedate Dates & times od the files
sub update_files
{
	my ($PID, $files, $filesize, $filedate) = @_;

	foreach (@$files)
	{
		$db->do ("UPDATE file SET DateAdded=UNIX_TIMESTAMP(),FileSize=$filesize->{$_},FileDate=". $db->quote ($filedate->{$_}) . " WHERE FileName=" . $db->quote ($_) . " AND PID=$PID");
	}
}

# update summary
sub update_status
{
	&log_print ("Updating status", 1);
	my $directories = $db->selectrow_array ("SELECT COUNT(*) FROM path");
	my ($files, $filesize) = $db->selectrow_array ("SELECT COUNT(*),SUM(FileSize) FROM file");
	my $flag_mask = $c{HOST_DELETE} | $c{HOST_PURGE};
	my $smb_hosts = $db->selectrow_array ("SELECT COUNT(*) FROM host WHERE HostType=$c{SMB} AND (Flags&$flag_mask)=0 AND TotalFileSize>=0");
	my $ftp_hosts = $db->selectrow_array ("SELECT COUNT(*) FROM host WHERE HostType=$c{FTP} AND (Flags&$flag_mask)=0 AND TotalFileSize>=0");

	my ($sec, $min, $hour, $day, $month, $year) = localtime time;
	my $last_change = $db->quote (sprintf ('%02d. %02d. %4d %02d:%02d', $day, $month + 1, $year + 1900, $hour, $min));

	$db->do ("LOCK TABLES status WRITE");
	my $queries = $db->selectrow_array ("SELECT Queries FROM status");
	$queries = 0 if (!defined ($queries));

	$db->do ("DELETE FROM status");
	$db->do ("INSERT INTO status (LastChange,SMBHosts,FTPHosts,Directories,Files,FileSize,Queries,UpdatingHost) VALUES ($last_change,$smb_hosts,$ftp_hosts,$directories,$files,$filesize,$queries,0)");
	$db->do ("UNLOCK TABLES");
}

# update total file size of the server
# \param $HID Host ID of the server
sub update_host_file_size
{
	my ($HID) = @_;

	my $filesize = $db->selectrow_array ("SELECT SUM(FileSize) FROM path LEFT JOIN file USING (PID) WHERE HID=$HID");
	$filesize = 0 if (!$filesize);
	$db->do ("UPDATE host SET TotalFileSize=$filesize WHERE HID=$HID");
}

# update total file size of the share
# \param $HID Host ID of the server
# \param $SID ID of the share
sub update_share_file_size
{
	my ($HID, $SID) = @_;

	my $filesize = $db->selectrow_array ("SELECT SUM(FileSize) FROM path LEFT JOIN file USING (PID) WHERE HID=$HID AND SID=$SID");
	$filesize = 0 if (!$filesize);
	$db->do ("UPDATE share SET TotalFileSize=$filesize WHERE HID=$HID AND SID=$SID");
}

# optimize storage of tables
sub optimize_tables
{
	my $tables = $db->selectcol_arrayref ("SHOW TABLES");
	if (!$tables)
	{
		&log_print ("Error listing tables - reason: $DBI::errstr\n", 0);
		return 1;
	}

	foreach (@$tables)
	{
		&log_print ("Optimizing table $_", 1);
		$db->do ("OPTIMIZE TABLE $_");
	}
}

# repair tables
sub repair_tables
{
	my $tables = $db->selectcol_arrayref ("SHOW TABLES");
	if (!$tables)
	{
		&log_print ("Error listing tables - reason: $DBI::errstr\n", 0);
		return 1;
	}

	foreach (@$tables)
	{
		&log_print ("Repairing table $_", 1);
		$db->do ("REPAIR TABLE $_");
	}
}

# fix inconsistencies in database
sub fix_tables
{
	&log_print ("Fixing inconsistencies.", 1);
	my ($HIDs, $SIDs, $PIDs);

	# delete servers from table host which do not have a record in appropriate table (smb, ftp)
	$HIDs = $db->selectcol_arrayref ("SELECT host.HID FROM host LEFT JOIN smb USING (HID) WHERE HostType=$c{SMB} AND smb.HID is NULL");
	foreach (@$HIDs)
	{
		&purge_host ($_);
	}
	$HIDs = $db->selectcol_arrayref ("SELECT host.HID FROM host LEFT JOIN ftp USING (HID) WHERE HostType=$c{FTP} AND ftp.HID is NULL");
	foreach (@$HIDs)
	{
		&purge_host ($_);
	}

	# delete servers which are not in table host
	$HIDs = $db->selectcol_arrayref ("SELECT smb.HID FROM smb LEFT JOIN host USING (HID) WHERE host.HID is NULL");
	foreach (@$HIDs)
	{
		&purge_host ($_);
	}
	$HIDs = $db->selectcol_arrayref ("SELECT ftp.HID FROM ftp LEFT JOIN host USING (HID) WHERE host.HID is NULL");
	foreach (@$HIDs)
	{
		&purge_host ($_);
	}

	# delete shares which are not on any server
	$SIDs = $db->selectcol_arrayref ("SELECT share.SID FROM share LEFT JOIN host USING (HID) WHERE host.HID is NULL");
	&purge_shares ($SIDs);

	# delete directories which are not on any server
	$PIDs = $db->selectcol_arrayref ("SELECT path.PID FROM path LEFT JOIN host USING (HID) WHERE host.HID is NULL");
	&purge_paths ($PIDs);

	# delete files which are not in any directory
	$PIDs = $db->selectcol_arrayref ("SELECT DISTINCT file.PID FROM file LEFT JOIN path USING (PID) WHERE path.PID is NULL");
	&purge_paths ($PIDs);

	&log_print ("Updating total file size for servers and shares.", 1);

	$HIDs = $db->selectcol_arrayref ("SELECT HID FROM host WHERE HostType=$c{SMB} AND TotalFileSize>=0");
	foreach my $HID (@$HIDs)
	{
		$SIDs = $db->selectcol_arrayref ("SELECT SID FROM share WHERE HID=$HID");
		foreach my $SID (@$SIDs)
		{
			&update_share_file_size ($HID, $SID);
		}
		&update_host_file_size ($HID);
	}
	$HIDs = $db->selectcol_arrayref ("SELECT HID FROM host WHERE HostType=$c{FTP} AND TotalFileSize>=0");
	foreach my $HID (@$HIDs)
	{
		&update_host_file_size ($HID);
	}
}
