#!/usr/bin/perl

# Add FTP servers from specified IPs to the list
# Copyright (c) 2002-2003 Josef Zlomek
#
# This file is part of Fast File Search (FFS).
#
# FFS is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA,
# or download it from http://www.gnu.org/licenses/gpl.html

use strict;
use DBI;
use Getopt::Long;
use Socket;
use Net::FTP;

# include the common package
$_ = $0;
$_ =~ s/[^\/]*$/common.pl/;
require;

my ($help, $private);
my $login = 'anonymous';
my $password = 'user@host.com';
my $port = 21;
my $passive = 1;
my $recursive = 1;
my $update = 0;
my $set_flags = 0;
my $clear_flags = 0;
Getopt::Long::Configure ("bundling", "no_ignore_case");
GetOptions ('help|h' => \$help,
		'login=s' => \$login,
		'password=s' => \$password,
		'port=i' => \$port,
		'passive' => \$passive,
		'active' => sub { $passive = 0 },
		'no-recursive' => sub { $recursive = 0 },
		'update|u' => \$update,
		'common|master|c' => sub { $private = 0 },
		'private|slave|p' => \$private,
		);
if ($help)
{
	&help ();
	exit 0;
}

my %c;
my $db;

&read_config (\%c);
if ($private)
{
	&db_connect (\%c, undef, \$db);
}
else
{
	&db_connect (\%c, \$db);
}
exit 1 if (!$db);

my $q_login = $db->quote ($login);
my $q_password = $db->quote ($password);
$port = int ($port);
$port = 21 if ($port <= 0 || $port > 65535);
if ($passive)
{
	$set_flags |= $c{HOST_PASSIVE_MODE};
}
else
{
	$clear_flags |= $c{HOST_PASSIVE_MODE};
}
if ($recursive)
{
	$set_flags |= $c{HOST_LS_RECURSIVE};
}
else
{
	$clear_flags |= $c{HOST_LS_RECURSIVE};
}

foreach (@ARGV)
{
	my @ip1 = ();
	my @ip2 = ();
	my @ip3 = ();
	my @ip4 = ();
	my ($i1,$i2,$i3,$i4);

	if (/^([\d,*-]+)\.([\d,*-]+)\.([\d,*-]+)\.([\d,*-]+)$/)
	{
		&set_ip_part (\@ip1, $1);
		&set_ip_part (\@ip2, $2);
		&set_ip_part (\@ip3, $3);
		&set_ip_part (\@ip4, $4);
	}
	elsif (/^(\d+)\.(\d+)\.(\d+)\.(\d+)\/(\d+)$/)
	{
		my $mask = 256 - (1 << (8 - $5 % 8));
		my $last = (1 << (8 - $5 % 8)) - 1;
		if ($5 >= 0 && $5 < 8)
		{
			&set_range (\@ip1, ($1 & $mask), ($1 & $mask) + $last);
			&set_range (\@ip2, 0, 255);
			&set_range (\@ip3, 0, 255);
			&set_range (\@ip4, 0, 255);
		}
		elsif ($5 >= 8 && $5 < 16)
		{
			$ip1[$1] = 1;
			&set_range (\@ip2, ($2 & $mask), ($2 & $mask) + $last);
			&set_range (\@ip3, 0, 255);
			&set_range (\@ip4, 0, 255);
		}
		elsif ($5 >= 16 && $5 < 24)
		{
			$ip1[$1] = 1;
			$ip2[$2] = 1;
			&set_range (\@ip3, ($3 & $mask), ($3 & $mask) + $last);
			&set_range (\@ip4, 0, 255);
		}
		elsif ($5 >= 24 && $5 < 32)
		{
			$ip1[$1] = 1;
			$ip2[$2] = 1;
			$ip3[$3] = 1;
			&set_range (\@ip4, ($4 & $mask), ($4 & $mask) + $last);
		}
		else
		{
			$ip1[$1] = 1;
			$ip2[$2] = 1;
			$ip3[$3] = 1;
			$ip4[$4] = 1;
		}
	}

	for ($i1 = 0; $i1 < 256; $i1++)
	{
		if ($ip1[$i1])
		{
			for ($i2 = 0; $i2 < 256; $i2++)
			{
				if ($ip2[$i2])
				{
					for ($i3 = 0; $i3 < 256; $i3++)
					{
						if ($ip3[$i3])
						{
							for ($i4 = 0; $i4 < 256; $i4++)
							{
								if ($ip4[$i4])
								{
									&process_ip ($i1, $i2, $i3, $i4);
								}
							}
						}
					}
				}
			}
		}
	}
}
&db_disconnect ($db);

sub set_ip_part
{
	my ($arr, $str) = @_;

	my @parts = split (/,/, $str);
	foreach (@parts)
	{
		$arr->[$1] = 1 if (/^(\d+)$/);
		&set_range ($arr, 0, 255) if ($_ eq "*");
		&set_range ($arr, $1, $2) if (/^(\d+)-(\d+)$/);
	}
}

sub set_range
{
	my ($arr, $x, $y) = @_;

	for (; $x <= $y; $x++)
	{
		$arr->[$x] = 1;
	}
}

sub process_ip
{
	my @ip = @_;

	my ($ip_str, $name);
	my @host_ent;

	$ip_str = join (".", @ip);

	# try to connect to FTP server
	my $ftp_handle = Net::FTP->new ($ip_str, Port => $port, Timeout => $c{FTP_TIMEOUT}, Passive => $passive);
	return if (!defined $ftp_handle);
	if ($password eq 'user@host.com')
	{
		$password = $c{FTP_ANON_PASSWD};
	}
	if (!$ftp_handle->login ($login, $password))
	{
		$ftp_handle->close ();
		return;
	}
	$ftp_handle->close ();

	@host_ent = gethostbyaddr (pack ("C4", @ip), AF_INET);
	if (!@host_ent)
	{
		$name = $ip_str;
	}
	else
	{
		$name = uc ($host_ent[0]);
	}

	&update_host_list ($ip_str, $name);
}

sub update_host_list
{
	my ($ip, $name) = @_;

	$db->do ("LOCK TABLES host WRITE, ftp WRITE");
	my ($resHID, $resflags) = $db->selectrow_array ("SELECT host.HID,Flags FROM ftp LEFT JOIN host USING (HID) WHERE HostName=" . $db->quote ($name));
	if ($resHID)
	{
		if ($update)
		{
			$db->do ("UPDATE host SET IP=" . $db->quote ($ip) . ",Flags=(Flags|$set_flags)&~$clear_flags WHERE HID=$resHID");
			$db->do ("UPDATE ftp SET Login=$q_login,PassWord=$q_password,Port=$port WHERE HID=$resHID");
		}
		else
		{
			$db->do ("UPDATE host SET IP=" . $db->quote ($ip) . " WHERE HID=$resHID");
		}
	}
	else
	{
		$db->do ("INSERT INTO host (HostType,ExpireCount,Period,Counter,ScanOrder,Flags,IP) VALUES ($c{FTP},1,1,1,0,$set_flags," . $db->quote ($ip) . ")");
		$db->do ("INSERT INTO ftp (HID,HostName,Login,PassWord,Port) VALUES (LAST_INSERT_ID()," . $db->quote ($name) . ",$q_login,$q_password,$port)");
	}
	$db->do ("UNLOCK TABLES");
}

sub help
{
	print "Fast File Search ftp_add_IPs.pl - Add FTP servers from specified IPs to the list\n";
	print "Usage: ftp_add_IPs.pl [OPTION]... [IP]...\n\n";
	print "You can use several formats to specify IP blocks, the following IP blocks are equivalent:\n";
	print "  192.168.*.*\n";
	print "  192.168.0.0/16\n";
	print "  192.168.0-64,65-255.0,1,2,3-255\n";
	print "\n";
	print "        --login=LOGIN        FTP login (default: anonymous)\n";
	print "        --password=PASSWORD  FTP password (default: user\@host.com)\n";
	print "        --port=PORT          FTP port (default: 21)\n";
	print "        --passive            Use passive mode for FTP data transfers (default)\n";
	print "        --active             Use active mode for FTP data transfers\n";
	print "        --no-recursive       Never use recursive ls (ls -lR) for getting file list\n";
	print "        --update             Update information (Login, Password, ...) about server\n";
	print "                               if server already is in database\n";
	print "    -c, --common, --master   Use common (master) database (default)\n";
	print "    -p, --private, --slave   Use private (slave) database\n";
	print "    -h, --help               Display this help and exit\n\n";
}
