#!/usr/bin/perl

# Create listing of files from database
# Copyright (c) 2002-2003 Josef Zlomek
#
# This file is part of Fast File Search (FFS).
#
# FFS is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA,
# or download it from http://www.gnu.org/licenses/gpl.html

use strict;
use DBI;
use Getopt::Long;

# include the common package
$_ = $0;
$_ =~ s/[^\/]*$/common.pl/;
require;

# prefix of file list name (e.g. directory that file lists will be created in)
my $prefix = "./";

# shall the output be sorted?
my $sort = 0;

# types of file lists that will be generated
# syntax: [ file_name, [ list of masks ], min. file size, max. file size ]
# example: [ "example1.txt", [ "*.ext1" ], "50M", "1G" ],
# example: [ "example2.txt", [ "*.ext2", "*.ext3" ], "30m" ],
my @filelist = (
	[ 'all.txt', [ '*' ] ],
	[ 'iso.txt', [ '*.iso'] ],
	[ 'linux.txt', [ 'linux-2.*' ] ],
	[ 'music.txt', [ '**.mp3', '*.ogg'] ],
	[ 'pictures.txt', [ '*.jpg', '*.jpeg', '*.gif', '*.png' ] ],
	[ 'software.txt', [ '*.rpm', '*.deb', '*.gz', '*.bz2' ] ]
);

my $help;
Getopt::Long::Configure ("bundling", "no_ignore_case");
GetOptions ('help|h' => \$help,
		'prefix|p=s' => \$prefix,
		'sort|s' => \$sort,
		);
if ($help)
{
	&help ();
	exit 0;
}

my %c;
my $db;

&read_config (\%c);
&db_connect (\%c, undef, \$db);
exit 1 if (!$db);

# get host list
my @hosts;
my $where;
my $unset_flags = $c{HOST_DELETE} | $c{HOST_PURGE};
if ($c{SMB_DISABLE} eq 'no')
{
	my $hosts = $db->selectall_arrayref ("SELECT smb.HID,HostName FROM smb LEFT JOIN host USING (HID) WHERE Flags&$unset_flags=0");
	foreach (@$hosts)
	{
		push @hosts, [ $_->[0], "file://" . $_->[1], $_->[1] ];
	}
}
if ($c{FTP_DISABLE} eq 'no')
{
	my $hosts = $db->selectall_arrayref ("SELECT ftp.HID,HostName,Login,PassWord,Port FROM ftp LEFT JOIN host USING (HID) WHERE Flags&$unset_flags=0");
	foreach (@$hosts)
	{
		my $str = "ftp://";
		$str .= $_->[2] . ':' . $_->[3] . '@' if ($c{SHOW_LOGIN_PASS} eq 'yes' && ($_->[2] ne 'anonymous' || $_->[3] ne 'user@host.com'));
		$str .= $_->[1];
		$str .= ':' . $_->[4] if ($_->[4] != 21);
		push @hosts, [ $_->[0], $str, $_->[1] ];
	}
}
# host names may be always sorted, because there are not too many hosts so it is fast
@hosts = sort { $a->[2] cmp $b->[2] } @hosts;

# write the file lists
foreach my $file (@filelist)
{
	if (open (FO, ">$prefix" . $file->[0]))
	{
		$where = "(";
		foreach (@{$file->[1]})
		{
			my $str = $_;
			my $end = substr (reverse ($_), 0, 4);
			$end =~ s/\*.*/*/;
			$str =~ s/([%_])/\\$1/g;
			$end =~ s/([%_])/\\$1/g;
			$str =~ y/*?/%_/;
			$end =~ y/*?/%_/;
			$where .= " OR " if ($where ne "(");
			$where .= "FileName LIKE " . $db->quote ($str) . " AND FileNameEnd LIKE " . $db->quote ($end);
		}
		$where .= ")";

		my ($min, $max);
		$min = &get_size ($file->[2], 0);
		$max = &get_size ($file->[3], 0xffffffff);
		$where .= " AND FileSize>=$min" if ($min > 0);
		$where .= " AND FileSize<=$max" if ($max < 0xffffffff);

		if ($sort)
		{
			foreach (@hosts)
			{
				&list_host ($_->[0], $_->[1]);
			}
		}
		else
		{
			foreach (@hosts)
			{
				&list_host_unsorted ($_->[0], $_->[1]);
			}
		}
		close (FO);
	}
	else
	{
		print "Can't write to $prefix", $file->[0], "\n";
	}
}

&db_disconnect ($db);

sub list_host
{
	my ($HID, $str) = @_;

	my $cmd = $db->prepare ("SELECT PID,FullName FROM path WHERE HID=$HID ORDER BY FullName");
	$cmd->execute ();
	while (my @row = $cmd->fetchrow_array ())
	{
		&list_path ($row[0], $str . $row[1]);
	}
}

sub list_path
{
	my ($PID, $str) = @_;

	my $cmd = $db->prepare ("SELECT FileName,FileSize FROM file WHERE PID=$PID AND $where ORDER BY FileName");
	$cmd->execute ();
	while (my @row = $cmd->fetchrow_array ())
	{
		printf FO "%10d  ", $row[1];
		print FO "$str/$row[0]\r\n";
	}
}

sub list_host_unsorted
{
	my ($HID, $str) = @_;

	my $cmd = $db->prepare ("SELECT FullName,FileName,FileSize FROM path LEFT JOIN file USING (PID) WHERE path.HID=$HID AND $where");
	$cmd->execute ();
	while (my @row = $cmd->fetchrow_array ())
	{
		printf FO "%10d  ", $row[2];
		print FO "$str$row[0]/$row[1]\r\n";
	}
}

sub help
{
	print "Fast File Search list.pl - Create listing of files from database\n";
	print "Usage: list.pl [OPTION]...\n\n";
	print "    -p, --prefix=PREFIX   Set the PREFIX of names of created files\n";
	print "    -s, --sort            Sort the output in alphabetical order\n";
	print "    -h, --help            Display this help and exit\n\n";
}
