#!/usr/bin/perl

# Check whether the hosts in database are online and set the flags
# Copyright (c) 2003 Josef Zlomek
#
# This file is part of Fast File Search (FFS).
#
# FFS is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA,
# or download it from http://www.gnu.org/licenses/gpl.html

use strict;
use DBI;
use Getopt::Long;
use Socket;
use Fcntl;
use IO::Poll qw(POLLIN POLLOUT POLLHUP POLLERR);

# include the common package
$_ = $0;
$_ =~ s/[^\/]*$/common.pl/;
require;
use vars qw($import_num $import_nnum $import_str0 $import_str1);

my %c;
my $db;

my ($help);
Getopt::Long::Configure ("bundling", "no_ignore_case");
GetOptions ('help|h' => \$help);
if ($help)
{
	&help ();
	exit 0;
}
&read_config (\%c);
&db_connect (\%c, undef, \$db);
exit 1 if (!$db);

my $hosts;

# mark the hosts which are deleted or do not have IP address in database as NOT ONLINE
$db->do ("UPDATE host SET Flags=Flags&~" . $c{HOST_ONLINE} . " WHERE IP='' OR Flags&" . ($c{HOST_DELETE} | $c{HOST_PURGE}) . ">0");

if ($c{ONLINE_CHECK_SMB} eq 'nmblookup')
{
	$hosts = $db->selectall_arrayref ("SELECT HID,IP FROM host WHERE HostType='$c{SMB}' AND IP!='' AND Flags&" . ($c{HOST_DELETE} | $c{HOST_PURGE}) . "=0");
	&online_check_nmblookup ($hosts);
}
elsif ($c{ONLINE_CHECK_SMB} eq 'ping')
{
	$hosts = $db->selectall_arrayref ("SELECT HID,IP FROM host WHERE HostType='$c{SMB}' AND IP!='' AND Flags&" . ($c{HOST_DELETE} | $c{HOST_PURGE}) . "=0");
	&online_check_ping ($hosts, $c{ONLINE_CHECK_SMB_TIMEOUT});
}
elsif ($c{ONLINE_CHECK_SMB} eq 'connect')
{
	$hosts = $db->selectall_arrayref ("SELECT host.HID,IP,139 FROM smb LEFT JOIN host USING (HID) WHERE IP!='' AND Flags&" . ($c{HOST_DELETE} | $c{HOST_PURGE}) . "=0");
	&online_check_connect ($hosts, $c{ONLINE_CHECK_SMB_TIMEOUT}, $c{ONLINE_CHECK_SMB_PARALLEL});
}

if ($c{ONLINE_CHECK_FTP} eq 'ping')
{
	$hosts = $db->selectall_arrayref ("SELECT HID,IP FROM host WHERE HostType='$c{FTP}' AND IP!='' AND Flags&" . ($c{HOST_DELETE} | $c{HOST_PURGE}) . "=0");
	&online_check_ping ($hosts, $c{ONLINE_CHECK_FTP_TIMEOUT});
}
elsif ($c{ONLINE_CHECK_FTP} eq 'connect')
{
	$hosts = $db->selectall_arrayref ("SELECT host.HID,IP,Port FROM ftp LEFT JOIN host USING (HID) WHERE IP!='' AND Flags&" . ($c{HOST_DELETE} | $c{HOST_PURGE}) . "=0");
	&online_check_connect ($hosts, $c{ONLINE_CHECK_FTP_TIMEOUT}, $c{ONLINE_CHECK_FTP_PARALLEL});
}
&db_disconnect ($db);

sub online_check_nmblookup
{
	my ($hosts) = @_;

	if (!open (NMBLOOKUP, $c{SMB_PATH} . "/nmblookup '*' |"))
	{
		die ("Can't run \"nmblookup '*'\".\n");
	}

	my %ip = map { chomp; s/ .*//; $_ => "\n" } <NMBLOOKUP>;
	close (NMBLOOKUP);

	my $host;
	foreach $host (@$hosts)
	{
		if ($ip{$host->[1]})
		{
			$db->do ("UPDATE host SET Flags=Flags|" . $c{HOST_ONLINE} . " WHERE HID=" . $host->[0]);
		}
		else
		{
			$db->do ("UPDATE host SET Flags=Flags&~" . $c{HOST_ONLINE} . " WHERE HID=" . $host->[0]);
		}
	}
}

sub online_check_ping
{
	my ($hosts, $timeout) = @_;

	my $host;
	foreach $host (@$hosts)
	{
		my $out = `ping -c1 -i1 -w$timeout $host->[1] 2>/dev/null`;
		if ($out =~ /time=/)
		{
			$db->do ("UPDATE host SET Flags=Flags|" . $c{HOST_ONLINE} . " WHERE HID=" . $host->[0]);
		}
		else
		{
			$db->do ("UPDATE host SET Flags=Flags&~" . $c{HOST_ONLINE} . " WHERE HID=" . $host->[0]);
		}
	}
}

sub online_check_connect
{
	my ($hosts, $timeout, $parallel) = @_;

	my @group;
	my $host;
	foreach $host (@$hosts)
	{
		$host->[3] = &connect_nonblock ($host->[1], $host->[2]);
		push @group, $host;
		if (@group == $parallel)
		{
			&check_group (\@group, $timeout);
			@group = ();
		}
	}
	&check_group (\@group, $timeout) if (@group > 0);
}

sub connect_nonblock
{
	my ($host, $port) = @_;
	local *SOCK;

	my $iaddr = inet_aton ($host);
	my $paddr = sockaddr_in ($port, $iaddr);
	socket (SOCK, PF_INET, SOCK_STREAM, getprotobyname ('tcp'));
	fcntl (SOCK, F_SETFL, O_NONBLOCK);
	connect (SOCK, $paddr);

	return *SOCK;
}

sub check_group
{
	my ($group, $timeout) = @_;

	my ($host, $set, $clear);
	my $poll = new IO::Poll;

	foreach $host (@$group)
	{
		$poll->mask ($host->[3] => POLLIN | POLLOUT | POLLHUP | POLLERR);
	}

	select (undef, undef, undef, $timeout);	# wait some time
	$poll->poll (0);		# test which sockets are connected

	foreach $host (@$group)
	{
		my $events = $poll->events ($host->[3]);
		if (($events & (POLLIN | POLLOUT)) && !($events & (POLLHUP | POLLERR)))
		{
			$set .= " OR " if ($set);
			$set .= "HID=$host->[0]";
		}
		else
		{
			$clear .= " OR " if ($clear);
			$clear .= "HID=$host->[0]";
		}
		close ($host->[3]);
	}
	$db->do ("UPDATE host SET Flags=Flags|" . $c{HOST_ONLINE} . " WHERE $set") if ($set);
	$db->do ("UPDATE host SET Flags=Flags&~" . $c{HOST_ONLINE} . " WHERE $clear") if ($clear);
}

sub help
{
	print "Fast File Search online_check.pl - Check whether servers are online\n";
	print "Usage: online_check.pl [OPTION]...\n\n";
	print "    -h, --help               Display this help and exit\n\n";
}
