#!/usr/bin/perl

# Add SMB servers from specified IPs to the list
# Copyright (c) 2002-2003 Josef Zlomek
#
# This file is part of Fast File Search (FFS).
#
# FFS is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA,
# or download it from http://www.gnu.org/licenses/gpl.html

use strict;
use DBI;
use Getopt::Long;

# include the common package
$_ = $0;
$_ =~ s/[^\/]*$/common.pl/;
require;

my ($help, $private);
my $params = 1024;
my $broadcast = 0;
Getopt::Long::Configure ("bundling", "no_ignore_case");
GetOptions ('help|h' => \$help,
	    	'broadcast|b' => \$broadcast,
		'params=i' => \$params,
		'common|master|c' => sub { $private = 0 },
		'private|slave|p' => \$private,
		);
if ($help)
{
	&help ();
	exit 0;
}
$params = 1 if ($params < 1);

my %c;
my $db;

&read_config (\%c);
if ($private)
{
	&db_connect (\%c, undef, \$db);
}
else
{
	&db_connect (\%c, \$db);
}
exit 1 if (!$db);

if ($broadcast)
{
	if (@ARGV == 0)
	{
		&process_broadcast ("");
	}
	else
	{
		foreach (@ARGV)
		{
			next if (&is_invalid_IP ($_));
			&process_broadcast ("-B $_");
		}
	}
}
else
{
	my @ip_list;
	foreach (@ARGV)
	{
		my @ip1 = ();
		my @ip2 = ();
		my @ip3 = ();
		my @ip4 = ();
		my ($i1,$i2,$i3,$i4);

		if (/^([\d,*-]+)\.([\d,*-]+)\.([\d,*-]+)\.([\d,*-]+)$/)
		{
			&set_ip_part (\@ip1, $1);
			&set_ip_part (\@ip2, $2);
			&set_ip_part (\@ip3, $3);
			&set_ip_part (\@ip4, $4);
		}
		elsif (/^(\d+)\.(\d+)\.(\d+)\.(\d+)\/(\d+)$/)
		{
			my $mask = 256 - (1 << (8 - $5 % 8));
			my $last = (1 << (8 - $5 % 8)) - 1;
			if ($5 >= 0 && $5 < 8)
			{
				&set_range (\@ip1, ($1 & $mask), ($1 & $mask) + $last);
				&set_range (\@ip2, 0, 255);
				&set_range (\@ip3, 0, 255);
				&set_range (\@ip4, 0, 255);
			}
			elsif ($5 >= 8 && $5 < 16)
			{
				$ip1[$1] = 1;
				&set_range (\@ip2, ($2 & $mask), ($2 & $mask) + $last);
				&set_range (\@ip3, 0, 255);
				&set_range (\@ip4, 0, 255);
			}
			elsif ($5 >= 16 && $5 < 24)
			{
				$ip1[$1] = 1;
				$ip2[$2] = 1;
				&set_range (\@ip3, ($3 & $mask), ($3 & $mask) + $last);
				&set_range (\@ip4, 0, 255);
			}
			elsif ($5 >= 24 && $5 < 32)
			{
				$ip1[$1] = 1;
				$ip2[$2] = 1;
				$ip3[$3] = 1;
				&set_range (\@ip4, ($4 & $mask), ($4 & $mask) + $last);
			}
			else
			{
				$ip1[$1] = 1;
				$ip2[$2] = 1;
				$ip3[$3] = 1;
				$ip4[$4] = 1;
			}
		}

		for ($i1 = 0; $i1 < 256; $i1++)
		{
			if ($ip1[$i1])
			{
				for ($i2 = 0; $i2 < 256; $i2++)
				{
					if ($ip2[$i2])
					{
						for ($i3 = 0; $i3 < 256; $i3++)
						{
							if ($ip3[$i3])
							{
								for ($i4 = 0; $i4 < 256; $i4++)
								{
									if ($ip4[$i4])
									{
										push (@ip_list, "$i1.$i2.$i3.$i4");
										if (@ip_list == $params)
										{
											&process_list (\@ip_list);
											@ip_list = ();
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
	&process_list (\@ip_list) if (@ip_list > 0);
}
&db_disconnect ($db);

sub set_ip_part
{
	my ($arr, $str) = @_;

	my @parts = split (/,/, $str);
	foreach (@parts)
	{
		$arr->[$1] = 1 if (/^(\d+)$/);
		&set_range ($arr, 0, 255) if ($_ eq "*");
		&set_range ($arr, $1, $2) if (/^(\d+)-(\d+)$/);
	}
}

sub set_range
{
	my ($arr, $x, $y) = @_;

	for (; $x <= $y; $x++)
	{
		$arr->[$x] = 1;
	}
}

sub process_broadcast
{
	my ($param) = @_;

	my @output = `$c{SMB_PATH}/nmblookup $c{SMB_CONF} $param '*'`;
	my @ip_list;
	my $i;
	for ($i = 0; $i < @output; $i++)
	{
		if ($output[$i] =~ /^(\d+\.\d+\.\d+\.\d+)/)
		{
			push (@ip_list, $1);
			if (@ip_list == $params)
			{
				&process_list (\@ip_list);
				@ip_list = ();
			}

		}
	}
	&process_list (\@ip_list) if (@ip_list > 0);
}

sub process_list
{
	my ($list) = @_;

	my @output = `$c{SMB_PATH}/nmblookup $c{SMB_CONF} -A @$list`;
	my @worklist;
	my $i;
	for ($i = 0; $i < @output; $i++)
	{
		if ($output[$i] =~ /(\d+\.\d+\.\d+\.\d+)$/)
		{
			my $ip = $1;
			my $host;
			my $group;
			for ($i++; $output[$i] !~ /^\r?\n$/; $i++)
			{
				if (substr ($output[$i], 18, 2) eq "00")
				{
					if (substr ($output[$i], 25, 5) eq "GROUP")
					{
						$group = substr ($output[$i], 1, 15);
						$group =~ s/\s+$//;
					}
					else
					{
						$host = substr ($output[$i], 1, 15);
						$host =~ s/\s+$//;
					}
				}
			}
			push (@worklist, [ $ip, $host, $group ] ) if ($host);
		}
	}
	&update_host_list (\@worklist);
}

sub update_host_list
{
	my ($worklist) = @_;

	foreach (@$worklist)
	{
		$db->do ("LOCK TABLES host WRITE, smb WRITE");
		my ($resHID, $resflags) = $db->selectrow_array ("SELECT host.HID,Flags FROM smb LEFT JOIN host USING (HID) WHERE HostName=" . $db->quote ($_->[1]));
		if ($resHID)
		{
			$db->do ("UPDATE host SET IP=" . $db->quote ($_->[0]) . " WHERE HID=$resHID");
			$db->do ("UPDATE smb SET WorkGroup=" . $db->quote ($_->[2]) . " WHERE HID=$resHID");
		}
		else
		{
			$db->do ("INSERT INTO host (HostType,ExpireCount,Period,Counter,ScanOrder,Flags,IP) VALUES ($c{SMB},1,1,1,0,$c{HOST_MANUAL_CHANGES}," . $db->quote ($_->[0]) . ")");
			$db->do ("INSERT INTO smb (HID,HostName,WorkGroup) VALUES (LAST_INSERT_ID()," . $db->quote ($_->[1]) . "," . $db->quote ($_->[2]) . ")");
		}
		$db->do ("UNLOCK TABLES");
	}
}

sub help
{
	print "Fast File Search smb_add_IPs.pl - Add SMB servers from specified IPs to the list\n";
	print "Usage: smb_add_IPs.pl [OPTION]... [IP_BLOCK]...\n";
	print "       smb_add_IPs.pl --broadcast [OPTION]... [BROADCAST_IP]...\n";
	print "\n";
	print "You can use several formats to specify IP blocks, the following IP blocks are equivalent:\n";
	print "  192.168.*.*\n";
	print "  192.168.0.0/16\n";
	print "  192.168.0-64,65-255.0,1,2,3-255\n";
	print "\n";
	print "        --broadcast, -b      Use broadcast to find servers\n";
	print "        --params=PARAMS      Number of parameteres passed to nmblookup\n";
	print "                               (default = 1024)\n";
	print "    -c, --common, --master   Use common (master) database (default)\n";
	print "    -p, --private, --slave   Use private (slave) database\n";
	print "    -h, --help               Display this help and exit\n\n";
}
